import os
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path

from rich import print

from erc7730.common.json import dict_from_json_file, dict_to_json_file
from erc7730.common.output import (
    AddFileOutputAdder,
    BufferAdder,
    ConsoleOutputAdder,
    DropFileOutputAdder,
    ExceptionsToOutput,
    OutputAdder,
)
from erc7730.list.list import get_erc7730_files


def format_all_and_print_errors(paths: list[Path]) -> bool:
    """
    Format all ERC-7730 descriptor files at given paths and print errors.

    :param paths: paths to apply formatter on
    :return: true if not errors occurred
    """
    out = DropFileOutputAdder(delegate=ConsoleOutputAdder())

    count = format_all(paths, out)

    if out.has_errors:
        print(f"[bold][red]formatted {count} descriptor files, some errors occurred ❌[/red][/bold]")
        return False

    if out.has_warnings:
        print(f"[bold][yellow]formatted {count} descriptor files, some warnings occurred ⚠️[/yellow][/bold]")
        return True

    print(f"[bold][green]formatted {count} descriptor files, no errors occurred ✅[/green][/bold]")
    return True


def format_all(paths: list[Path], out: OutputAdder) -> int:
    """
    Format all ERC-7730 descriptor files at given paths.

    Paths can be files or directories, in which case all descriptor files in the directory are recursively formatted.

    :param paths: paths to apply formatter on
    :param out: output adder
    :return: number of files formatted
    """
    files = list(get_erc7730_files(*paths, out=out))

    if len(files) <= 1 or not (root_path := os.path.commonpath(files)):
        root_path = None

    def label(f: Path) -> Path | None:
        return f.relative_to(root_path) if root_path is not None else None

    if len(files) > 1:
        print(f"📝 formatting {len(files)} descriptor files…\n")

    with ThreadPoolExecutor() as executor:
        for future in (executor.submit(format_file, file, out, label(file)) for file in files):
            future.result()

    return len(files)


def format_file(path: Path, out: OutputAdder, show_as: Path | None = None) -> None:
    """
    Format a single ERC-7730 descriptor file.

    :param path: ERC-7730 descriptor file path
    :param show_as: if provided, print this label instead of the file path
    :param out: error handler
    """

    label = path if show_as is None else show_as
    file_out = AddFileOutputAdder(delegate=out, file=path)

    with BufferAdder(file_out, prolog=f"➡️ formatting [bold]{label}[/bold]…", epilog="") as out, ExceptionsToOutput(out):
        dict_to_json_file(path, dict_from_json_file(path))



"""Ledger specific utilities."""


def ledger_network_id(chain_id: int) -> str | None:
    """Get Ledger specific network id from chain id."""
    match chain_id:
        case 1:
            return "ethereum"
        case 10:
            return "optimism"
        case 100:
            return "gnosis"
        case 100:
            return "palm_network"
        case 10001:
            return "ethereum_pow"
        case 1001:
            return "klaytn_baobab"
        case 106:
            return "velas_evm"
        case 108:
            return "thundercore"
        case 1088:
            return "metis"
        case 1101:
            return "polygon_zk_evm"
        case 11155111:
            return "ethereum_sepolia"
        case 11155420:
            return "optimism_sepolia"
        case 1284:
            return "moonbeam"
        case 1285:
            return "moonriver"
        case 1313114:
            return "ether1"
        case 137:
            return "polygon"
        case 14:
            return "flare"
        case 1442:
            return "polygon_zk_evm_testnet"
        case 16:
            return "flare_coston"
        case 1620:
            return "atheios"
        case 1666600000:
            return "harmony_one"
        case 168587773:
            return "blast_sepolia"
        case 17000:
            return "ethereum_holesky"
        case 19:
            return "songbird"
        case 196:
            return "okbchain"
        case 199:
            return "bittorrent"
        case 199:
            return "bttc"
        case 2:
            return "expanse"
        case 200625:
            return "akroma"
        case 20531812:
            return "tecracoin"
        case 2221:
            return "kava_evm_testnet"
        case 2222:
            return "kava_evm"
        case 237:
            return "dexon"
        case 24:
            return "kardia_chain"
        case 24484:
            return "webchain"
        case 245022934:
            return "neon_evm"
        case 246:
            return "energy_web"
        case 246529:
            return "artis_sigma1"
        case 25:
            return "cronos"
        case 250:
            return "fantom"
        case 269:
            return "hpb"
        case 288:
            return "boba"
        case 2894:
            return "reosc"
        case 3:
            return "ethereum_ropsten"
        case 30:
            return "rsk"
        case 31102:
            return "ethergem"
        case 31102:
            return "ethersocial"
        case 3125659152:
            return "pirl"
        case 314:
            return "filecoin"
        case 314159:
            return "filecoin_calibration"
        case 321:
            return "kcc"
        case 324:
            return "zksync"
        case 336:
            return "shiden_evm"
        case 338:
            return "cronos_testnet"
        case 369:
            return "pulsechain"
        case 39797:
            return "energi"
        case 40:
            return "telos_evm"
        case 4002:
            return "fantom_testnet"
        case 42:
            return "lukso"
        case 42:
            return "xdc"
        case 420:
            return "optimism_goerli"
        case 42161:
            return "arbitrum"
        case 421613:
            return "arbitrum_goerli"
        case 421614:
            return "arbitrum_sepolia"
        case 42220:
            return "celo"
        case 42793:
            return "etherlink"
        case 43113:
            return "avalanche_c_chain_fuji"
        case 43114:
            return "avalanche_c_chain"
        case 44787:
            return "celo_alfajores"
        case 4919:
            return "venidium"
        case 5:
            return "ethereum_goerli"
        case 5000:
            return "mantle"
        case 5003:
            return "mantle_sepolia"
        case 51:
            return "apothem_network"
        case 534351:
            return "scroll_sepolia"
        case 534352:
            return "scroll"
        case 56:
            return "bsc"
        case 57:
            return "syscoin"
        case 59144:
            return "linea"
        case 592:
            return "astar"
        case 60:
            return "gochain"
        case 61:
            return "ethereum_classic"
        case 62621:
            return "multivac"
        case 64:
            return "ellaism"
        case 66:
            return "okxchain"
        case 76:
            return "mix"
        case 7762959:
            return "musicoin"
        case 8:
            return "ubiq"
        case 80001:
            return "polygon_mumbai"
        case 81457:
            return "blast"
        case 820:
            return "callisto"
        case 8217:
            return "klaytn"
        case 8453:
            return "base"
        case 84531:
            return "base_goerli"
        case 84532:
            return "base_sepolia"
        case 846000:
            return "id4good"
        case 88:
            return "tomo"
        case 888:
            return "wanchain"
        case 89:
            return "tomo_testnet"
        case 9001:
            return "evmos_evm"
        case 99:
            return "poa"
        case _:
            return None



"""Package implementing an object model for ERC-7730 descriptors."""

from enum import StrEnum, auto


class ERC7730ModelType(StrEnum):
    """
    The type of the ERC-7730 model.
    """

    INPUT = auto()
    RESOLVED = auto()



import json
from typing import Any

from pydantic import TypeAdapter

from erc7730.common.options import first_not_none
from erc7730.common.output import OutputAdder
from erc7730.common.pydantic import model_to_json_str
from erc7730.convert.resolved.constants import ConstantProvider
from erc7730.convert.resolved.parameters import resolve_field_parameters
from erc7730.convert.resolved.values import resolve_field_value
from erc7730.model.display import (
    FieldFormat,
)
from erc7730.model.input.display import (
    InputFieldDefinition,
    InputFieldParameters,
    InputReference,
)
from erc7730.model.metadata import EnumDefinition
from erc7730.model.paths import DataPath, DescriptorPath, Field
from erc7730.model.paths.path_ops import descriptor_path_strip_prefix
from erc7730.model.resolved.display import (
    ResolvedField,
    ResolvedFieldDescription,
    ResolvedFieldParameters,
)
from erc7730.model.types import Id

DEFINITIONS_PATH = DescriptorPath(elements=[Field(identifier="display"), Field(identifier="definitions")])


def resolve_reference(
    prefix: DataPath,
    reference: InputReference,
    definitions: dict[Id, InputFieldDefinition],
    enums: dict[Id, EnumDefinition],
    constants: ConstantProvider,
    out: OutputAdder,
) -> ResolvedField | None:
    if (definition := _get_definition(reference.ref, definitions, out)) is None:
        return None

    if (label := first_not_none(reference.label, definition.label)) is None:
        return out.error(
            title="Missing display field label",
            message=f"Label must be defined either on display field, or on the referenced display field definition "
            f"{reference.ref}.",
        )

    params: dict[str, Any] = {}
    if (definition_params := definition.params) is not None:
        params.update(json.loads(model_to_json_str(definition_params)))
    if (reference_params := reference.params) is not None:
        params.update(reference_params)

    resolved_params: ResolvedFieldParameters | None = None

    if params:
        input_params: InputFieldParameters = TypeAdapter(InputFieldParameters).validate_json(json.dumps(params))
        if (resolved_params := resolve_field_parameters(prefix, input_params, enums, constants, out)) is None:
            return None

    if (value := resolve_field_value(prefix, reference, definition.format, constants, out)) is None:
        return None

    return ResolvedFieldDescription(
        value=value,
        label=str(constants.resolve(label, out)),
        format=FieldFormat(definition.format),
        params=resolved_params,
    )


def _get_definition(
    ref: DescriptorPath, definitions: dict[Id, InputFieldDefinition], out: OutputAdder
) -> InputFieldDefinition | None:
    if (definition_id := _get_definition_id(ref, out)) is None:
        return None

    if (definition := definitions.get(definition_id)) is None:
        return out.error(
            title="Invalid display definition reference",
            message=f"""Display definition "{definition_id}" does not exist, valid ones are: """
            f"{', '.join(definitions.keys())}.",
        )
    return definition


def _get_definition_id(ref: DescriptorPath, out: OutputAdder) -> Id | None:
    try:
        tail = descriptor_path_strip_prefix(ref, DEFINITIONS_PATH)
    except ValueError:
        return out.error(
            title="Invalid definition reference path",
            message=f"References to display field definitions are restricted to {DEFINITIONS_PATH}, {ref} "
            f"cannot be used as a field definition reference.",
        )
    if len(tail.elements) != 1:
        return out.error(
            title="Invalid definition reference path",
            message=f"References to display field definitions are restricted to fields immediately under "
            f"{DEFINITIONS_PATH}, deep nesting is not allowed, {ref} cannot be used as a field "
            f"definition reference.",
        )
    if not isinstance(element := tail.elements[0], Field):
        return out.error(
            title="Invalid definition reference path",
            message=f"References to display field definitions are restricted to fields immediately under "
            f"{DEFINITIONS_PATH}, array operators are not allowed, {ref} cannot be used as a field "
            f"definition reference.",
        )

    return element.identifier



import pytest
from pydantic import TypeAdapter
from pydantic_string_url import HttpUrl

from erc7730.common.output import RaisingOutputAdder
from erc7730.convert.resolved.constants import DefaultConstantProvider
from erc7730.model.input.context import InputContract, InputContractContext, InputDeployment
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.input.display import InputDisplay, InputFieldDescription, InputFormat
from erc7730.model.input.metadata import InputMetadata
from erc7730.model.input.path import DescriptorPathStr
from erc7730.model.paths import DescriptorPath
from erc7730.model.paths.path_parser import to_path


def _provider(**constants: str | int | bool | float | None) -> DefaultConstantProvider:
    return DefaultConstantProvider(
        InputERC7730Descriptor(
            context=InputContractContext(
                contract=InputContract(
                    abi=HttpUrl("https://example.net/abi.json"),
                    deployments=[
                        InputDeployment(chainId=1, address="0x1111111111111111111111111111111111111111"),
                        InputDeployment(chainId=42, address="0x4242424242424242424242424242424242424242"),
                    ],
                )
            ),
            metadata=InputMetadata(constants=constants),
            display=InputDisplay(
                formats={"test": InputFormat(fields=[InputFieldDescription(path=to_path("#.foo"), label="Foo")])}
            ),
        )
    )


def _descriptor_path(value: str) -> DescriptorPath:
    return TypeAdapter(DescriptorPathStr).validate_strings(value)


def test_get_string() -> None:
    assert _provider(foo="bar").get(_descriptor_path("$.metadata.constants.foo"), RaisingOutputAdder()) == "bar"


def test_get_bool() -> None:
    assert _provider(foo=False).get(_descriptor_path("$.metadata.constants.foo"), RaisingOutputAdder()) is False


def test_get_int() -> None:
    assert _provider(foo=42).get(_descriptor_path("$.metadata.constants.foo"), RaisingOutputAdder()) == 42


def test_get_float() -> None:
    assert _provider(foo=1.42).get(_descriptor_path("$.metadata.constants.foo"), RaisingOutputAdder()) == 1.42


def test_get_from_context() -> None:
    assert _provider().get(_descriptor_path("$.context.contract.abi"), RaisingOutputAdder()) == HttpUrl(
        "https://example.net/abi.json"
    )


def test_get_from_list() -> None:
    assert _provider().get(_descriptor_path("$.context.contract.deployments.[1].chainId"), RaisingOutputAdder()) == 42


def test_resolve_literal() -> None:
    assert _provider(foo="bar").resolve("baz", RaisingOutputAdder()) == "baz"


def test_resolve_constant() -> None:
    assert _provider(foo="bar").resolve(to_path("$.metadata.constants.foo"), RaisingOutputAdder()) == "bar"


def test_resolve_or_none_literal() -> None:
    assert _provider(foo="bar").resolve_or_none("baz", RaisingOutputAdder()) == "baz"


def test_resolve_or_none_constant() -> None:
    assert _provider(foo="bar").resolve_or_none(to_path("$.metadata.constants.foo"), RaisingOutputAdder()) == "bar"


def test_resolve_or_none_none() -> None:
    assert _provider(foo="bar").resolve_or_none(None, RaisingOutputAdder()) is None


def test_resolve_path_literal_data_path() -> None:
    assert _provider(foo="bar").resolve_path(to_path("#.a.b.c"), RaisingOutputAdder()) == to_path("#.a.b.c")


def test_resolve_path_literal_container_path() -> None:
    assert _provider(foo="bar").resolve_path(to_path("@.to"), RaisingOutputAdder()) == to_path("@.to")


def test_resolve_path_constant() -> None:
    assert _provider(foo="#.a.b.c").resolve_path(to_path("$.metadata.constants.foo"), RaisingOutputAdder()) == to_path(
        "#.a.b.c"
    )


def test_resolve_path_or_none_literal_data_path() -> None:
    assert _provider(foo="bar").resolve_path_or_none(to_path("#.a.b.c"), RaisingOutputAdder()) == to_path("#.a.b.c")


def test_resolve_path_or_none_literal_container_path() -> None:
    assert _provider(foo="bar").resolve_path_or_none(to_path("@.to"), RaisingOutputAdder()) == to_path("@.to")


def test_resolve_path_or_none_constant() -> None:
    assert _provider(foo="#.a.b.c").resolve_path_or_none(
        to_path("$.metadata.constants.foo"), RaisingOutputAdder()
    ) == to_path("#.a.b.c")


def test_resolve_path_or_none_none() -> None:
    assert _provider(foo="#.a.b.c").resolve_path_or_none(None, RaisingOutputAdder()) is None


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_get_invalid_no_such_constant() -> None:
    _provider(foo="bar").get(_descriptor_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_get_invalid_empty_constants() -> None:
    _provider().get(_descriptor_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match=".*Path \\$.metadata.\\[0\\] is invalid, \\$.metadata is not an array.")
def test_get_invalid_not_an_array() -> None:
    _provider(foo="bar").get(_descriptor_path("$.metadata.[0]"), RaisingOutputAdder())


@pytest.mark.raises(match=".*\\$.context.contract.deployments is an array.")
def test_get_invalid_not_a_dict() -> None:
    _provider(foo="bar").get(_descriptor_path("$.context.contract.deployments.foo"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_resolve_invalid_no_such_constant() -> None:
    _provider(foo="bar").resolve(to_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_resolve_invalid_empty_constants() -> None:
    _provider().resolve(to_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_resolve_path_invalid_no_such_constant() -> None:
    _provider(foo="bar").resolve_path(to_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_resolve_path_invalid_empty_constants() -> None:
    _provider().resolve_path(to_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_resolve_path_or_none_invalid_no_such_constant() -> None:
    _provider(foo="bar").resolve_path_or_none(to_path("$.metadata.constants.baz"), RaisingOutputAdder())


@pytest.mark.raises(match='.*\\$.metadata.constants has no "baz" field.')
def test_resolve_path_or_none_invalid_empty_constants() -> None:
    _provider().resolve_path_or_none(to_path("$.metadata.constants.baz"), RaisingOutputAdder())



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from pydantic import BaseModel

from erc7730.common.output import OutputAdder

InputType = TypeVar("InputType", bound=BaseModel)
OutputType = TypeVar("OutputType", bound=BaseModel)


class ERC7730Converter(ABC, Generic[InputType, OutputType]):
    """
    Converter from/to ERC-7730 descriptor.

    A converter may fail partially, in which case it should emit errors with ERROR level, or totally, in which case it
    should emit errors with FATAL level.
    """

    @abstractmethod
    def convert(self, descriptor: InputType, out: OutputAdder) -> OutputType | dict[str, OutputType] | None:
        """
        Convert a descriptor from/to ERC-7730.

        Conversion may fail partially, in which case it should emit errors with WARNING level, or totally, in which case
        it should emit errors with ERROR level.

        Conversion can return a single descriptor, or multiple ones, in the form of a dictionary with unique
        identifiers.

        :param descriptor: input descriptor to convert
        :param out: output sink
        :return: converted descriptor, or None if conversion failed
        """
        raise NotImplementedError()



import pytest

from erc7730.common.abi import (
    compute_signature,
    reduce_signature,
    signature_to_selector,
)
from erc7730.model.abi import Component, Function, InputOutput


@pytest.mark.parametrize(
    "signature, expected",
    [
        # no param, no space
        ("transfer()", "transfer()"),
        # one param, no space
        ("transfer(address)", "transfer(address)"),
        # multiple params, expected spaces, no names
        ("mintToken(uint256, uint256, address, uint256, bytes)", "mintToken(uint256,uint256,address,uint256,bytes)"),
        # multiple params, expected spaces, names
        (
            "mintToken(uint256 eventId, uint256 tokenId, address receiver, uint256 expirationTime, bytes signature)",
            "mintToken(uint256,uint256,address,uint256,bytes)",
        ),
        # multiple params, spaces everywhere, names, end with tuple
        (
            "f1( uint256[] _a , address _o , ( uint256 v , uint256 d ) _p )",
            "f1(uint256[],address,(uint256,uint256))",
        ),
        # multiple params, spaces everywhere, names, tuple in middle
        (
            "f2( uint256[] _a , ( uint256 v , uint256 d ) _p , address _o )",
            "f2(uint256[],(uint256,uint256),address)",
        ),
        # multiple params, spaces everywhere, names, nested tuples
        (
            "f3( uint256[] _a , ( uint256 v , ( bytes s , address a ) , uint256 d ) _p , address _o )",
            "f3(uint256[],(uint256,(bytes,address),uint256),address)",
        ),
    ],
)
def test_reduce_signature(signature: str, expected: str) -> None:
    # This test covers both parse_signature and compute_signature
    assert reduce_signature(signature) == expected


def test_compute_signature_no_params() -> None:
    abi = Function(name="transfer", inputs=[])
    expected = "transfer()"
    assert compute_signature(abi) == expected


def test_compute_signature_with_params() -> None:
    abi = Function(
        name="transfer", inputs=[InputOutput(name="to", type="address"), InputOutput(name="amount", type="uint256")]
    )
    expected = "transfer(address,uint256)"
    assert compute_signature(abi) == expected


def test_compute_signature_with_nested_params() -> None:
    abi = Function(
        name="foo",
        inputs=[
            InputOutput(
                name="bar",
                type="tuple",
                components=[Component(name="baz", type="uint256"), Component(name="qux", type="address")],
            )
        ],
    )
    expected = "foo((uint256,address))"
    assert compute_signature(abi) == expected


def test_signature_to_selector() -> None:
    signature = "transfer(address,uint256)"
    expected = "0xa9059cbb"
    assert signature_to_selector(signature) == expected



from typing import assert_never

from erc7730.model.paths import (
    ROOT_DATA_PATH,
    ContainerPath,
    DataPath,
    DataPathElement,
    DescriptorPath,
    DescriptorPathElement,
)


def descriptor_path_strip_prefix(path: DescriptorPath, prefix: DescriptorPath) -> DescriptorPath:
    """
    Strip expected prefix from a descriptor path, raising an error if the prefix is not matching.

    :param path: path to strip
    :param prefix: prefix to strip
    :return: path without prefix
    :raises ValueError: if the path does not start with the prefix
    """
    if len(path.elements) < len(prefix.elements):
        raise ValueError(f"Path {path} does not start with prefix {prefix}.")
    for i, element in enumerate(prefix.elements):
        if path.elements[i] != element:
            raise ValueError(f"Path {path} does not start with prefix {prefix}.")
    return DescriptorPath(elements=path.elements[len(prefix.elements) :])


def data_path_strip_prefix(path: DataPath, prefix: DataPath) -> DataPath:
    """
    Strip expected prefix from a data path, raising an error if the prefix is not matching.

    :param path: path to strip
    :param prefix: prefix to strip
    :return: path without prefix
    :raises ValueError: if the path does not start with the prefix
    """
    if path.absolute != prefix.absolute or len(path.elements) < len(prefix.elements):
        raise ValueError(f"Path {path} does not start with prefix {prefix}.")
    for i, element in enumerate(prefix.elements):
        if path.elements[i] != element:
            raise ValueError(f"Path {path} does not start with prefix {prefix}.")
    return DataPath(absolute=path.absolute, elements=path.elements[len(prefix.elements) :])


def descriptor_path_starts_with(path: DescriptorPath, prefix: DescriptorPath) -> bool:
    """
    Check if descriptor path starts with a given prefix.

    :param path: path to inspect
    :param prefix: prefix to check
    :return: True if path starts with prefix
    """
    try:
        descriptor_path_strip_prefix(path, prefix)
        return True
    except ValueError:
        return False


def data_path_starts_with(path: DataPath, prefix: DataPath) -> bool:
    """
    Check if data path starts with a given prefix.

    :param path: path to inspect
    :param prefix: prefix to check
    :return: True if path starts with prefix
    """
    try:
        data_path_strip_prefix(path, prefix)
        return True
    except ValueError:
        return False


def path_starts_with(
    path: DataPath | ContainerPath | DescriptorPath, prefix: DataPath | ContainerPath | DescriptorPath
) -> bool:
    """
    Check if path starts with a given prefix.

    :param path: path to inspect
    :param prefix: prefix to check
    :return: True if path starts with prefix
    """
    match (path, prefix):
        case (ContainerPath(), ContainerPath()):
            return path == prefix
        case (DataPath(), DataPath()):
            return data_path_starts_with(path, prefix)
        case (DescriptorPath(), DescriptorPath()):
            return descriptor_path_starts_with(path, prefix)
        case _:
            return False


def descriptor_path_ends_with(path: DescriptorPath, suffix: DescriptorPathElement) -> bool:
    """
    Check if descriptor path ends with a given element.

    :param path: path to inspect
    :param suffix: suffix to check
    :return: True if path ends with suffix
    """
    return path.elements[-1] == suffix


def data_path_ends_with(path: DataPath, suffix: DataPathElement) -> bool:
    """
    Check if data path ends with a given element.

    :param path: path to inspect
    :param suffix: suffix to check
    :return: True if path ends with suffix
    """
    if not path.elements:
        return False
    return path.elements[-1] == suffix


def data_path_concat(parent: DataPath | None, child: DataPath) -> DataPath:
    """
    Concatenate two data paths.

    :param parent: parent path
    :param child: child path
    :return: concatenated path
    """
    if parent is None or child.absolute:
        return child
    return DataPath(absolute=parent.absolute, elements=[*parent.elements, *child.elements])


def data_or_container_path_concat(parent: DataPath | None, child: DataPath | ContainerPath) -> DataPath | ContainerPath:
    """
    Concatenate a data path with either another data path, or a container path.

    :param parent: parent path
    :param child: child path
    :return: concatenated path
    """
    if parent is None:
        return child
    match child:
        case DataPath() as child:
            return data_path_concat(parent, child)
        case ContainerPath() as child:
            return child
        case _:
            assert_never(child)


def data_path_append(parent: DataPath, child: DataPathElement) -> DataPath:
    """
    Append an element to a data path.

    :param parent: parent path
    :param child: element to append
    :return: concatenated path
    """
    return parent.model_copy(update={"elements": [*parent.elements, child]})


def descriptor_path_append(parent: DescriptorPath, child: DescriptorPathElement) -> DescriptorPath:
    """
    Append an element to a descriptor path.

    :param parent: parent path
    :param child: element to append
    :return: concatenated path
    """
    return parent.model_copy(update={"elements": [*parent.elements, child]})


def to_absolute(path: DataPath | ContainerPath) -> DataPath | ContainerPath:
    """
    Convert a path to an absolute path.

    :param path: data path
    :return: absolute path
    """
    match path:
        case DataPath():
            return data_path_concat(ROOT_DATA_PATH, path)
        case ContainerPath():
            return path
        case _:
            assert_never(path)


def to_relative(path: DataPath | ContainerPath) -> DataPath | ContainerPath:
    """
    Convert a path to a relative path.

    :param path: data path
    :return: absolute path
    """
    match path:
        case DataPath():
            return path.model_copy(update={"absolute": False})
        case ContainerPath():
            return path
        case _:
            assert_never(path)



import json
from pathlib import Path

import pytest

from erc7730.common.json import read_json_with_includes
from erc7730.model.input.descriptor import InputERC7730Descriptor
from tests.assertions import assert_dict_equals
from tests.cases import path_id
from tests.files import ERC7730_DESCRIPTORS
from tests.schemas import assert_valid_erc_7730


@pytest.mark.parametrize("input_file", ERC7730_DESCRIPTORS, ids=path_id)
def test_schema(input_file: Path) -> None:
    """Test model serializes to JSON that matches the schema."""

    # TODO: invalid files in registry
    if input_file.name in {"eip712-rarible-erc-1155.json", "eip712-rarible-erc-721.json"}:
        pytest.skip("Rarible EIP-712 schemas are missing EIP712Domain")

    assert_valid_erc_7730(InputERC7730Descriptor.load(input_file))


@pytest.mark.parametrize("input_file", ERC7730_DESCRIPTORS, ids=path_id)
def test_round_trip(input_file: Path) -> None:
    """Test model serializes back to same JSON."""
    actual = json.loads(InputERC7730Descriptor.load(input_file).to_json_string())
    expected = read_json_with_includes(input_file)
    assert_dict_equals(expected, actual)



import re
import threading
from builtins import print as builtin_print
from contextlib import AbstractContextManager
from enum import IntEnum, auto
from itertools import groupby
from types import TracebackType
from typing import assert_never, final, override

from pydantic import BaseModel, ConfigDict, FilePath, ValidationError
from pydantic_core import ErrorDetails
from rich import print

MUX = threading.Lock()


class Output(BaseModel):
    """An output info/debug/warning/error."""

    model_config = ConfigDict(
        strict=True,
        frozen=True,
        extra="forbid",
        validate_default=True,
        validate_return=True,
        validate_assignment=True,
        arbitrary_types_allowed=False,
        allow_inf_nan=False,
    )

    class Level(IntEnum):
        """ERC7730Linter output level."""

        DEBUG = auto()
        INFO = auto()
        WARNING = auto()
        ERROR = auto()

    file: FilePath | None
    line: int | None
    title: str | None
    message: str
    level: Level = Level.ERROR


class OutputAdder:
    """An output debug/info/warning/error sink."""

    def __init__(self) -> None:
        self.has_infos = False
        self.has_warnings = False
        self.has_errors = False

    def add(self, output: Output) -> None:
        match output.level:
            case Output.Level.DEBUG:
                pass
            case Output.Level.INFO:
                self.has_infos = True
            case Output.Level.WARNING:
                self.has_warnings = True
            case Output.Level.ERROR:
                self.has_errors = True
            case _:
                assert_never(output.level)

    @final
    def debug(
        self, message: str, file: FilePath | None = None, line: int | None = None, title: str | None = None
    ) -> None:
        self.add(Output(file=file, line=line, title=title, message=message, level=Output.Level.DEBUG))

    @final
    def info(
        self, message: str, file: FilePath | None = None, line: int | None = None, title: str | None = None
    ) -> None:
        self.add(Output(file=file, line=line, title=title, message=message, level=Output.Level.INFO))

    @final
    def warning(
        self, message: str, file: FilePath | None = None, line: int | None = None, title: str | None = None
    ) -> None:
        self.add(Output(file=file, line=line, title=title, message=message, level=Output.Level.WARNING))

    @final
    def error(
        self, message: str, file: FilePath | None = None, line: int | None = None, title: str | None = None
    ) -> None:
        self.add(Output(file=file, line=line, title=title, message=message, level=Output.Level.ERROR))


@final
class ListOutputAdder(OutputAdder):
    """An output adder that stores outputs in a list."""

    def __init__(self) -> None:
        super().__init__()
        self.outputs: list[Output] = []

    def add(self, output: Output) -> None:
        super().add(output)
        self.outputs.append(output)


@final
class SetOutputAdder(OutputAdder):
    """An output adder that stores outputs in a set."""

    def __init__(self) -> None:
        super().__init__()
        self.outputs: set[Output] = set()

    def add(self, output: Output) -> None:
        super().add(output)
        self.outputs.add(output)


class ConsoleOutputAdder(OutputAdder):
    """An output adder that prints to the console."""

    @override
    def add(self, output: Output) -> None:
        super().add(output)
        match output.level:
            case Output.Level.DEBUG:
                style = "italic"
                prefix = "⚪️ "
            case Output.Level.INFO:
                style = "blue"
                prefix = "🔵 "
            case Output.Level.WARNING:
                style = "yellow"
                prefix = "🟠 warning: "
            case Output.Level.ERROR:
                style = "red"
                prefix = "🔴 error: "
            case _:
                assert_never(output.level)

        log = f"[{style}]{prefix}"
        context = []
        if output.file is not None:
            context.append(f"{output.file}")
        if output.line is not None:
            context.append(f"line {output.line}")
        if context:
            log += ", ".join(context) + ": "
        if output.title is not None:
            log += f"{output.title}: "
        log += f"[/{style}]"
        if "\n" in output.message:
            log += "\n"
        log += output.message

        print(log)


class RaisingOutputAdder(ConsoleOutputAdder):
    """An output adder that raises warnings/errors, otherwise prints to the console."""

    @override
    def add(self, output: Output) -> None:
        super().add(output)
        match output.level:
            case Output.Level.DEBUG | Output.Level.INFO:
                super().add(output)
            case Output.Level.WARNING | Output.Level.ERROR:
                log = f"{output.level.name}: "
                if output.file is not None:
                    log += f"file={output.file.name}"
                if output.line is not None:
                    log += f"line {output.line}: "
                if output.title is not None:
                    log += f"{output.title}: "
                log += f"{output.message}"
                raise Exception(log)
            case _:
                assert_never(output.level)


@final
class GithubAnnotationsAdder(OutputAdder):
    """An output adder that formats errors to be parsed as Github annotations."""

    @override
    def add(self, output: Output) -> None:
        super().add(output)

        match output.level:
            case Output.Level.DEBUG:
                return
            case Output.Level.INFO:
                lvl = "notice"
            case Output.Level.WARNING:
                lvl = "warning"
            case Output.Level.ERROR:
                lvl = "error"
            case _:
                assert_never(output.level)

        log = f"::{lvl} "
        if output.file is not None:
            log += f"file={output.file}"
        if output.line is not None:
            log += f",line={output.line}"
        if output.title is not None:
            log += f",title={output.title}"
        message = output.message.replace("\n", "%0A")
        log += f"::{message}"

        builtin_print(log)


class AddFileOutputAdder(OutputAdder):
    """An output adder wrapper that adds a specific file to all outputs."""

    def __init__(self, delegate: OutputAdder, file: FilePath) -> None:
        super().__init__()
        self.delegate: OutputAdder = delegate
        self.file: FilePath = file

    @override
    def add(self, output: Output) -> None:
        super().add(output)
        self.delegate.add(output.model_copy(update={"file": self.file}))


class DropFileOutputAdder(OutputAdder):
    """An output adder wrapper that drops file information from all outputs."""

    def __init__(self, delegate: OutputAdder) -> None:
        super().__init__()
        self.delegate: OutputAdder = delegate

    @override
    def add(self, output: Output) -> None:
        super().add(output)
        self.delegate.add(output.model_copy(update={"file": None}))


@final
class BufferAdder(AbstractContextManager[OutputAdder]):
    """A context manager that buffers outputs and outputs them all at once, sorted and deduplicated."""

    def __init__(self, delegate: OutputAdder, prolog: str | None = None, epilog: str | None = None) -> None:
        self._buffer = SetOutputAdder()
        self._delegate = delegate
        self._prolog = prolog
        self._epilog = epilog

    @override
    def __enter__(self) -> OutputAdder:
        return self._buffer

    @override
    def __exit__(self, etype: type[BaseException] | None, e: BaseException | None, tb: TracebackType | None) -> None:
        MUX.acquire()
        try:
            if self._prolog is not None:
                print(self._prolog)
            if self._buffer.outputs:
                for output in sorted(self._buffer.outputs, key=lambda x: (x.file, x.line, x.level, x.title, x.message)):
                    self._delegate.add(output)
            else:
                print("no issue found ✔️")
            if self._epilog is not None:
                print(self._epilog)
        finally:
            MUX.release()
        return None


@final
class ExceptionsToOutput(AbstractContextManager[None]):
    """A context manager that catches exceptions and redirects them to an OutputAdder."""

    def __init__(self, delegate: OutputAdder) -> None:
        self._delegate = delegate

    @override
    def __enter__(self) -> None:
        return None

    @override
    def __exit__(self, etype: type[BaseException] | None, e: BaseException | None, tb: TracebackType | None) -> bool:
        if isinstance(e, Exception):
            exception_to_output(e, self._delegate)
            return True
        return False


def exception_to_output(e: Exception, out: OutputAdder) -> None:
    """
    Sanitize an exception and add it to an OutputAdder.

    :param e: exception to handle
    :param out: output handler
    """
    match e:
        case ValidationError() as e:
            pydantic_error_to_output(e, out)
        case Exception() as e:
            out.error(title="Failed processing descriptor", message=str(e))
        case _:
            assert_never(e)


def pydantic_error_to_output(e: ValidationError, out: OutputAdder) -> None:
    """
    Sanitize a pydantic validation exception and add it to an OutputAdder.

    This cleans up location, and groups errors by location to avoid outputting multiple errors when not necessary, for
    instance for union types.

    :param e: exception to handle
    :param out: output handler
    """

    def filter_location(loc: int | str) -> bool:
        if isinstance(loc, int):
            return True
        return bool(re.match(r"(list|set)\[.*", loc))

    def get_location(ex: ErrorDetails) -> str:
        if not (loc := ex.get("loc")):
            return "unknown location"
        return ".".join(map(str, filter(filter_location, loc[:-1])))

    def get_value(ex: ErrorDetails) -> str:
        return str(ex.get("input", "unknown value"))

    def get_details(ex: ErrorDetails) -> str:
        return ex.get("msg", "unknown error")

    def get_message(ex: ErrorDetails) -> str:
        return f"""Value "{get_value(ex)}" is not valid: {get_details(ex)}"""

    for location, location_errors in groupby(e.errors(include_url=False), get_location):
        if (len(errors := list(location_errors))) > 1:
            out.error(title=f"Invalid value at {location}", message="* " + "\n * ".join(map(get_message, errors)))
        else:
            out.error(title=f"Invalid value at {location}", message=get_message(errors[0]))



"""
Object model for ERC-7730 descriptors `metadata` section.

Specification: https://github.com/LedgerHQ/clear-signing-erc7730-registry/tree/master/specs
JSON schema: https://github.com/LedgerHQ/clear-signing-erc7730-registry/blob/master/specs/erc7730-v1.schema.json
"""

from pydantic import Field
from pydantic_string_url import HttpUrl

from erc7730.model.metadata import Metadata
from erc7730.model.resolved.metadata import EnumDefinition
from erc7730.model.types import Id, ScalarType

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class InputMetadata(Metadata):
    """
    Metadata Section.

    The metadata section contains information about constant values relevant in the scope of the current contract /
    message (as matched by the `context` section)
    """

    constants: dict[Id, ScalarType | None] | None = Field(
        default=None,
        title="Constant values",
        description="A set of values that can be used in format parameters. Can be referenced with a path expression "
        "like $.metadata.constants.CONSTANT_NAME",
        examples=[
            {
                "token_path": "#.params.witness.outputs[0].token",
                "native_currency": "0x0000000000000000000000000000000000000001",
                "max_threshold": "0xFFFFFFFF",
                "max_message": "Max",
            }
        ],
    )

    enums: dict[Id, HttpUrl | EnumDefinition] | None = Field(
        default=None,
        title="Enums",
        description="A set of enums that are used to format fields replacing values with human readable strings.",
        examples=[{"interestRateMode": {"1": "stable", "2": "variable"}, "vaultIDs": "https://example.com/vaultIDs"}],
        max_length=32,  # TODO refine
    )



import os
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path

from rich import print

from erc7730.common.output import (
    AddFileOutputAdder,
    BufferAdder,
    ConsoleOutputAdder,
    DropFileOutputAdder,
    ExceptionsToOutput,
    GithubAnnotationsAdder,
    OutputAdder,
)
from erc7730.convert.resolved.convert_erc7730_input_to_resolved import ERC7730InputToResolved
from erc7730.lint import ERC7730Linter
from erc7730.lint.lint_base import MultiLinter
from erc7730.lint.lint_transaction_type_classifier import ClassifyTransactionTypeLinter
from erc7730.lint.lint_validate_abi import ValidateABILinter
from erc7730.lint.lint_validate_display_fields import ValidateDisplayFieldsLinter
from erc7730.list.list import get_erc7730_files
from erc7730.model.input.descriptor import InputERC7730Descriptor


def lint_all_and_print_errors(paths: list[Path], gha: bool = False) -> bool:
    out = GithubAnnotationsAdder() if gha else DropFileOutputAdder(delegate=ConsoleOutputAdder())

    count = lint_all(paths, out)

    if out.has_errors:
        print(f"[bold][red]checked {count} descriptor files, some errors found ❌[/red][/bold]")
        return False

    if out.has_warnings:
        print(f"[bold][yellow]checked {count} descriptor files, some warnings found ⚠️[/yellow][/bold]")
        return True

    print(f"[bold][green]checked {count} descriptor files, no errors found ✅[/green][/bold]")
    return True


def lint_all(paths: list[Path], out: OutputAdder) -> int:
    """
    Lint all ERC-7730 descriptor files at given paths.

    Paths can be files or directories, in which case all JSON files in the directory are recursively linted.

    :param paths: paths to apply linter on
    :param out: output adder
    :return: number of files checked
    """
    linter = MultiLinter(
        [
            ValidateABILinter(),
            ValidateDisplayFieldsLinter(),
            ClassifyTransactionTypeLinter(),
        ]
    )

    files = list(get_erc7730_files(*paths, out=out))

    if len(files) <= 1 or not (root_path := os.path.commonpath(files)):
        root_path = None

    def label(f: Path) -> Path | None:
        return f.relative_to(root_path) if root_path is not None else None

    if len(files) > 1:
        print(f"🔍 checking {len(files)} descriptor files…\n")

    with ThreadPoolExecutor() as executor:
        for future in (executor.submit(lint_file, file, linter, out, label(file)) for file in files):
            future.result()

    return len(files)


def lint_file(path: Path, linter: ERC7730Linter, out: OutputAdder, show_as: Path | None = None) -> None:
    """
    Lint a single ERC-7730 descriptor file.

    :param path: ERC-7730 descriptor file path
    :param show_as: if provided, print this label instead of the file path
    :param linter: linter instance
    :param out: error handler
    """

    label = path if show_as is None else show_as
    file_out = AddFileOutputAdder(delegate=out, file=path)

    with BufferAdder(file_out, prolog=f"➡️ checking [bold]{label}[/bold]…", epilog="") as out, ExceptionsToOutput(out):
        input_descriptor = InputERC7730Descriptor.load(path)
        resolved_descriptor = ERC7730InputToResolved().convert(input_descriptor, out)
        if resolved_descriptor is not None:
            linter.lint(resolved_descriptor, out)



from enum import StrEnum, auto
from typing import Annotated, Literal, Self, override

from pydantic import Field as PydanticField
from pydantic import (
    model_validator,
)

from erc7730.model.base import Model

ArrayIndex = Annotated[
    int,
    PydanticField(
        title="Array index",
        description="Index of an element in an array. An index can be negative to count from the end of the array.",
        ge=-32767,  # TODO to be refined
        le=32768,  # TODO to be refined
    ),
]


class Field(Model):
    """A path component designating a field in a structured data schema."""

    type: Literal["field"] = PydanticField(
        "field",
        title="Path Component Type",
        description="The path component type identifier (discriminator for path components discriminated union).",
    )

    identifier: str = PydanticField(
        title="Field Identifier",
        description="The identifier of the referenced field in the structured data schema.",
        pattern=r"^[a-zA-Z0-9_]+$",
    )

    @override
    def __str__(self) -> str:
        return self.identifier


class ArrayElement(Model):
    """A path component designating a single element of an array."""

    type: Literal["array_element"] = PydanticField(
        "array_element",
        title="Path Component Type",
        description="The path component type identifier (discriminator for path components discriminated union).",
    )

    index: ArrayIndex = PydanticField(
        title="Array Element",
        description="The index of the element in the array. It can be negative to count from the end of the array.",
    )

    @override
    def __str__(self) -> str:
        return f"[{self.index}]"


class ArraySlice(Model):
    """A path component designating an element range of an array (in which case, the path targets multiple values)."""

    type: Literal["array_slice"] = PydanticField(
        "array_slice",
        title="Path Component Type",
        description="The path component type identifier (discriminator for path components discriminated union).",
    )

    start: ArrayIndex | None = PydanticField(
        default=None,
        title="Slice Start Index",
        description="The start index of the slice (inclusive). Must be lower than the end index. If unset, slice "
        "starts from the beginning of the array.",
    )

    end: ArrayIndex | None = PydanticField(
        default=None,
        title="Slice End Index",
        description="The end index of the slice (exclusive). Must be greater than the start index. If unset, slice "
        "ends at the end of the array.",
    )

    @model_validator(mode="after")
    def _validate(self) -> Self:
        if (start := self.start) is None or (end := self.end) is None:
            return self
        if 0 <= end <= start or end <= start < 0:
            raise ValueError("Array slice start index must be strictly lower than end index.")
        return self

    def __str__(self) -> str:
        return f"[{'' if self.start is None else self.start}:{'' if self.end is None else self.end}]"


class Array(Model):
    """A path component designating all elements of an array (in which case, the path targets multiple values)."""

    type: Literal["array"] = PydanticField(
        "array",
        title="Path Component Type",
        description="The path component type identifier (discriminator for path components discriminated union).",
    )

    @override
    def __str__(self) -> str:
        return "[]"


class ContainerField(StrEnum):
    """
    Path applying to the container of the structured data to be signed.

    Such paths are prefixed with "@".
    """

    VALUE = auto()
    """The native currency value of the transaction containing the structured data."""

    FROM = auto()
    """The address of the sender of the transaction / signer of the message."""

    TO = auto()
    """The destination address of the containing transaction, ie the target smart contract address."""


DataPathElement = Annotated[
    Field | ArrayElement | ArraySlice | Array,
    PydanticField(
        title="Data Path Element",
        description="An element of a data path, applying to the structured data schema (ABI path for contracts, path"
        "in the message types itself for EIP-712)",
        discriminator="type",
    ),
]

DescriptorPathElement = Annotated[
    Field | ArrayElement,
    PydanticField(
        title="Descriptor Path Element",
        description="An element of a descriptor path, applying to the current file describing the structured data"
        "formatting, after merging with includes.",
        discriminator="type",
    ),
]


class ContainerPath(Model):
    """
    Path applying to the container of the structured data to be signed.

    Such paths are prefixed with "@".
    """

    type: Literal["container"] = PydanticField(
        "container",
        title="Path Type",
        description="The path type identifier (discriminator for paths discriminated union).",
    )

    field: ContainerField = PydanticField(
        title="Container field",
        description="The referenced field in the container, only some well-known values are allowed.",
    )

    @override
    def __str__(self) -> str:
        return f"@.{self.field}"


class DataPath(Model):
    """
    Path applying to the structured data schema (ABI path for contracts, path in the message types itself for
    EIP-712).

    A data path can reference multiple values if it contains array elements or slices.

    Such paths are prefixed with "#".
    """

    type: Literal["data"] = PydanticField(
        "data", title="Path Type", description="The path type identifier (discriminator for paths discriminated union)."
    )

    absolute: bool = PydanticField(
        title="Absolute",
        description="Whether the path is absolute (starting from the structured data root) or relative (starting from"
        "the current field).",
    )

    elements: list[DataPathElement] = PydanticField(
        title="Elements",
        description="The path elements, as a list of references to be interpreted left to right from the structured"
        "data root to reach the referenced value(s).",
    )

    @override
    def __str__(self) -> str:
        return f'{"#." if self.absolute else ""}{".".join(str(e) for e in self.elements)}'

    @override
    def __hash__(self) -> int:
        return hash(str(self))


class DescriptorPath(Model):
    """
    Path applying to the current file describing the structured data formatting, after merging with includes.

    A descriptor path can only reference a single value in the document.

    Such paths are prefixed with "$".
    """

    type: Literal["descriptor"] = PydanticField(
        "descriptor",
        title="Path Type",
        description="The path type identifier (discriminator for paths discriminated union).",
    )

    elements: list[DescriptorPathElement] = PydanticField(
        title="Elements",
        description="The path elements, as a list of references to be interpreted left to right from the current file"
        "root to reach the referenced value.",
    )

    @override
    def __str__(self) -> str:
        return f'$.{".".join(str(e) for e in self.elements)}'

    @override
    def __hash__(self) -> int:
        return hash(str(self))


ROOT_DATA_PATH = DataPath(absolute=True, elements=[])
ROOT_DESCRIPTOR_PATH = DescriptorPath(elements=[])



"""
Utilities for accessing ERC-7730 input descriptors nested fields.
"""

from erc7730.model.input.context import InputBindingContext, InputContractContext, InputDeployment, InputEIP712Context
from erc7730.model.input.descriptor import InputERC7730Descriptor


def get_chain_ids(descriptor: InputERC7730Descriptor) -> set[int]:
    """Get deployment chaind ids for a descriptor."""
    return {d.chainId for d in get_deployments(descriptor)}


def get_deployments(descriptor: InputERC7730Descriptor) -> list[InputDeployment]:
    """Get deployments section for a descriptor."""
    return get_binding_context(descriptor).deployments


def get_binding_context(descriptor: InputERC7730Descriptor) -> InputBindingContext:
    """Get binding context for a descriptor."""
    if isinstance(context := descriptor.context, InputEIP712Context):
        return context.eip712
    if isinstance(context := descriptor.context, InputContractContext):
        return context.contract
    raise ValueError(f"Invalid context type {type(descriptor.context)}")



from os import getcwd
from pathlib import Path

from erc7730 import (
    ERC_7730_REGISTRY_CALLDATA_PREFIX,
    ERC_7730_REGISTRY_DIRECTORY,
    ERC_7730_REGISTRY_EIP712_PREFIX,
)
from tests.io import load_json_file

# project root directory
PROJECT_ROOT = Path(getcwd())
while not (PROJECT_ROOT / "pyproject.toml").is_file():
    PROJECT_ROOT = PROJECT_ROOT.parent

# test resources
TEST_RESOURCES = PROJECT_ROOT / "tests" / "resources"
TEST_REGISTRIES = PROJECT_ROOT / "tests" / "registries"

# ERC-7730 registry resources
ERC7730_REGISTRY_ROOT = TEST_REGISTRIES / "clear-signing-erc7730-registry"
ERC7730_REGISTRY = ERC7730_REGISTRY_ROOT / ERC_7730_REGISTRY_DIRECTORY
ERC7730_CALLDATA_DESCRIPTORS = sorted(list(ERC7730_REGISTRY.rglob(f"{ERC_7730_REGISTRY_CALLDATA_PREFIX}*.json")))
ERC7730_EIP712_DESCRIPTORS = sorted(list(ERC7730_REGISTRY.rglob(f"{ERC_7730_REGISTRY_EIP712_PREFIX}*.json")))
ERC7730_DESCRIPTORS = sorted(ERC7730_CALLDATA_DESCRIPTORS + ERC7730_EIP712_DESCRIPTORS)
ERC7730_SCHEMA_PATH = ERC7730_REGISTRY_ROOT / "specs" / "erc7730-v1.schema.json"
ERC7730_SCHEMA = load_json_file(ERC7730_SCHEMA_PATH)

# legacy registry resources
LEGACY_REGISTRY = TEST_REGISTRIES / "ledger-asset-dapps"
LEGACY_EIP712_DESCRIPTORS = sorted(list(LEGACY_REGISTRY.rglob("**/eip712.json")))



from typing import final, override

from erc7730.lint.classifier import Classifier, TxClass
from erc7730.model.context import EIP712Schema


@final
class EIP712Classifier(Classifier[EIP712Schema]):
    """Given an EIP712 schema, classify the transaction type with some predefined ruleset.
    (implemented a basic detection of a permit)
    """

    @override
    def classify(self, schema: EIP712Schema) -> TxClass | None:
        if "permit" in schema.primaryType.lower():
            return TxClass.PERMIT
        return None



import pytest
from pydantic import TypeAdapter, ValidationError

from erc7730.model.input.path import ContainerPathStr, DataPathStr, DescriptorPathStr
from erc7730.model.paths import (
    Array,
    ArrayElement,
    ArraySlice,
    ContainerField,
    ContainerPath,
    DataPath,
    DescriptorPath,
    Field,
)
from erc7730.model.paths.path_parser import to_path
from erc7730.model.resolved.path import ResolvedPath
from tests.assertions import assert_json_str_equals


def _test_valid_input_path(string: str, obj: DescriptorPathStr | DataPathStr | ContainerPathStr, json: str) -> None:
    assert_json_str_equals(json, obj.to_json_string())
    assert to_path(string) == obj
    assert TypeAdapter(DescriptorPathStr | DataPathStr | ContainerPathStr).validate_json(f'"{string}"') == obj
    assert TypeAdapter(DescriptorPath | DataPath | ContainerPath).validate_json(json) == obj
    assert str(obj) == string


def _test_valid_resolved_path(string: str, obj: ResolvedPath, json: str) -> None:
    assert_json_str_equals(json, obj.to_json_string())
    assert to_path(string) == obj
    assert TypeAdapter(ResolvedPath).validate_json(json) == obj
    assert str(obj) == string


def test_valid_input_container_path() -> None:
    _test_valid_input_path(
        string="@.to",
        obj=ContainerPath(field=ContainerField.TO),
        json="""{ "type": "container", "field": "to" }""",
    )


def test_valid_input_data_path_absolute() -> None:
    _test_valid_input_path(
        string="#.params.[].[-2].[1:5].[:5].[5:].amountIn",
        obj=DataPath(
            absolute=True,
            elements=[
                Field(identifier="params"),
                Array(),
                ArrayElement(index=-2),
                ArraySlice(start=1, end=5),
                ArraySlice(start=None, end=5),
                ArraySlice(start=5, end=None),
                Field(identifier="amountIn"),
            ],
        ),
        json="""
            {
              "type": "data",
              "absolute": true,
              "elements": [
                { "type": "field", "identifier": "params" },
                { "type": "array" },
                { "type": "array_element", "index": -2 },
                { "type": "array_slice", "start": 1, "end": 5 },
                { "type": "array_slice", "end": 5 },
                { "type": "array_slice", "start": 5 },
                { "type": "field", "identifier": "amountIn" }
              ]
            }
        """,
    )


def test_valid_input_data_path_relative() -> None:
    _test_valid_input_path(
        string="params.[].[-2].[1:5].amountIn",
        obj=DataPath(
            absolute=False,
            elements=[
                Field(identifier="params"),
                Array(),
                ArrayElement(index=-2),
                ArraySlice(start=1, end=5),
                Field(identifier="amountIn"),
            ],
        ),
        json="""
            {
              "type": "data",
              "absolute": false,
              "elements": [
                { "type": "field", "identifier": "params" },
                { "type": "array" },
                { "type": "array_element", "index": -2 },
                { "type": "array_slice", "start": 1, "end": 5 },
                { "type": "field", "identifier": "amountIn" }
              ]
            }
        """,
    )


def test_valid_input_descriptor_path() -> None:
    _test_valid_input_path(
        string="$.params.[-2].[0].amountIn",
        obj=DescriptorPath(
            elements=[
                Field(identifier="params"),
                ArrayElement(index=-2),
                ArrayElement(index=0),
                Field(identifier="amountIn"),
            ]
        ),
        json="""
            {
              "type": "descriptor",
              "elements": [
                { "type": "field", "identifier": "params" },
                { "type": "array_element", "index": -2 },
                { "type": "array_element", "index": 0 },
                { "type": "field", "identifier": "amountIn" }
              ]
            }
        """,
    )


def test_valid_resolved_container_path() -> None:
    _test_valid_resolved_path(
        string="@.to",
        obj=ContainerPath(field=ContainerField.TO),
        json="""{ "type": "container", "field": "to" }""",
    )


def test_valid_resolved_data_path() -> None:
    _test_valid_resolved_path(
        string="#.params.[].[-2].[1:5].amountIn",
        obj=DataPath(
            absolute=True,
            elements=[
                Field(identifier="params"),
                Array(),
                ArrayElement(index=-2),
                ArraySlice(start=1, end=5),
                Field(identifier="amountIn"),
            ],
        ),
        json="""
            {
              "type": "data",
              "absolute": true,
              "elements": [
                { "type": "field", "identifier": "params" },
                { "type": "array" },
                { "type": "array_element", "index": -2 },
                { "type": "array_slice", "start": 1, "end": 5 },
                { "type": "field", "identifier": "amountIn" }
              ]
            }
        """,
    )


def test_invalid_container_value_unknown() -> None:
    with pytest.raises(ValueError) as e:
        to_path("@.foo")
    message = str(e.value)
    assert "Invalid path" in message
    assert "@.foo" in message
    assert "Expected one of" in message


def test_invalid_container_path_empty_component() -> None:
    with pytest.raises(ValueError) as e:
        to_path("@..foo")
    message = str(e.value)
    assert "Invalid path" in message
    assert "@..foo" in message
    assert "Expected one of" in message


def test_invalid_field_identifier_not_ascii() -> None:
    with pytest.raises(ValueError) as e:
        to_path("#.👿")
    message = str(e.value)
    assert "Invalid path" in message
    assert "#.👿" in message
    assert "Expected one of" in message


def test_invalid_array_element_index_not_a_number() -> None:
    with pytest.raises(ValueError) as e:
        to_path("#.[foo]")
    message = str(e.value)
    assert "Invalid path" in message
    assert "#.[foo]" in message
    assert "Expected one of" in message


def test_invalid_array_element_index_out_of_bounds() -> None:
    with pytest.raises(ValueError) as e:
        to_path("#.[65536]")
    message = str(e.value)
    assert "Invalid path" in message
    assert "#.[65536]" in message
    assert "Input should be less than or equal" in message


def test_invalid_array_slice_inverted() -> None:
    with pytest.raises(ValueError) as e:
        to_path("#.[1:0]")
    message = str(e.value)
    assert "Invalid path" in message
    assert "#.[1:0]" in message
    assert "Array slice start index must be strictly lower than end index" in message


def test_invalid_array_slice_used_in_descriptor_path() -> None:
    with pytest.raises(ValueError) as e:
        to_path("$.[1:0]")
    message = str(e.value)
    assert "Invalid path" in message
    assert "$.[1:0]" in message
    assert "Expected one of" in message


def test_invalid_array_used_in_descriptor_path() -> None:
    with pytest.raises(ValueError) as e:
        to_path("$.[]")
    message = str(e.value)
    assert "Invalid path" in message
    assert "$.[]" in message
    assert "Expected one of" in message


def test_invalid_relative_resolved_data_path() -> None:
    with pytest.raises(ValidationError) as e:
        TypeAdapter(ResolvedPath).validate_python(to_path("params.[].[-2].[1:5].amountIn"))
    message = str(e.value)
    assert "A resolved data path must be absolute" in message



from typing import final, override

from erc7730.lint.classifier import Classifier, TxClass
from erc7730.model.abi import ABI


@final
class ABIClassifier(Classifier[list[ABI]]):
    """Given an ABI, classify the transaction type with some predefined ruleset.
    (not implemented)
    """

    @override
    def classify(self, schema: list[ABI]) -> TxClass | None:
        pass



from typing import Annotated, Any, ForwardRef, Self

from pydantic import Discriminator, Field, Tag, model_validator

from erc7730.model.base import Model
from erc7730.model.display import (
    AddressNameType,
    DateEncoding,
    FieldFormat,
    FormatBase,
)
from erc7730.model.input.path import ContainerPathStr, DataPathStr, DescriptorPathStr
from erc7730.model.types import HexStr, Id, MixedCaseAddress, ScalarType
from erc7730.model.unions import field_discriminator, field_parameters_discriminator

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class InputFieldBase(Model):
    """
    A field formatter, containing formatting information of a single field in a message.
    """

    path: DescriptorPathStr | DataPathStr | ContainerPathStr | None = Field(
        default=None,
        title="Path",
        description="A path to the field in the structured data. The path is a JSON path expression that can be used "
        """to extract the field value from the structured data. Exactly one of "path" or "value" must be set.""",
    )

    value: DescriptorPathStr | ScalarType | None = Field(
        default=None,
        title="Value",
        description="A literal value on which the format should be applied instead of looking up a field in the "
        """structured data. Exactly one of "path" or "value" must be set.""",
    )

    @model_validator(mode="after")
    def _validate_one_of_path_or_value(self) -> Self:
        if self.path is None and self.value is None:
            raise ValueError('Either "path" or "value" must be set.')
        if self.path is not None and self.value is not None:
            raise ValueError('"path" and "value" are mutually exclusive.')
        return self


class InputReference(InputFieldBase):
    """
    A reference to a shared definition that should be used as the field formatting definition.

    The value is the key in the display definitions section, as a path expression $.display.definitions.DEFINITION_NAME.
    It is used to share definitions between multiple messages / functions.
    """

    ref: DescriptorPathStr = Field(
        alias="$ref",
        title="Internal Definition",
        description="An internal definition that should be used as the field formatting definition. The value is the "
        "key in the display definitions section, as a path expression $.display.definitions.DEFINITION_NAME.",
    )

    label: DescriptorPathStr | str | None = Field(
        default=None,
        title="Field Label",
        description="The label of the field, that will be displayed to the user in front of the formatted field value. "
        "Overrides the label in the referenced definition if set.",
    )

    params: dict[str, Any] | None = Field(
        default=None,
        title="Parameters",
        description="Parameters override. These values takes precedence over the ones in the definition itself.",
    )


class InputTokenAmountParameters(Model):
    """
    Token Amount Formatting Parameters.
    """

    tokenPath: DescriptorPathStr | DataPathStr | ContainerPathStr | None = Field(
        default=None,
        title="Token Path",
        description="Path reference to the address of the token contract. Used to associate correct ticker. If ticker "
        "is not found or tokenPath is not set, the wallet SHOULD display the raw value instead with an"
        '"Unknown token" warning. Exactly one of "tokenPath" or "token" must be set.',
    )

    token: DescriptorPathStr | MixedCaseAddress | None = Field(
        default=None,
        title="Token",
        description="The address of the token contract, as constant value. Used to associate correct ticker. If ticker "
        "is not found or value is not set, the wallet SHOULD display the raw value instead with an "
        '"Unknown token" warning. Exactly one of "tokenPath" or "token" must be set.',
    )

    nativeCurrencyAddress: list[DescriptorPathStr | MixedCaseAddress] | DescriptorPathStr | MixedCaseAddress | None = (
        Field(
            default=None,
            title="Native Currency Address",
            description="An address or array of addresses, any of which are interpreted as an amount in native "
            "currency rather than a token.",
        )
    )

    threshold: DescriptorPathStr | HexStr | int | None = Field(
        default=None,
        title="Unlimited Threshold",
        description="The threshold above which the amount should be displayed using the message parameter rather than "
        "the real amount (encoded as an int or byte array).",
    )

    message: DescriptorPathStr | str | None = Field(
        default=None,
        title="Unlimited Message",
        description="The message to display when the amount is above the threshold.",
    )

    @model_validator(mode="after")
    def _validate_one_of_token_path_or_value(self) -> Self:
        if self.tokenPath is not None and self.token is not None:
            raise ValueError('"tokenPath" and "token" are mutually exclusive.')
        return self


class InputAddressNameParameters(Model):
    """
    Address Names Formatting Parameters.
    """

    types: list[AddressNameType] | DescriptorPathStr | None = Field(
        default=None,
        title="Address Type",
        description="An array of expected types of the address. If set, the wallet SHOULD check that the address "
        "matches one of the types provided.",
        min_length=1,
    )

    sources: list[str] | DescriptorPathStr | None = Field(
        default=None,
        title="Trusted Sources",
        description="An array of acceptable sources for names. If set, the wallet SHOULD restrict name lookup to "
        "relevant sources.",
        min_length=1,
    )


class InputCallDataParameters(Model):
    """
    Embedded Calldata Formatting Parameters.
    """

    selector: DescriptorPathStr | str | None = Field(
        default=None,
        title="Called Selector",
        description="The selector being called, if not contained in the calldata. Hex string representation.",
    )

    calleePath: DescriptorPathStr | DataPathStr | ContainerPathStr | None = Field(
        default=None,
        title="Callee Path",
        description="The path to the address of the contract being called by this embedded calldata. Exactly one of "
        '"calleePath" or "callee" must be set.',
    )

    callee: DescriptorPathStr | MixedCaseAddress | None = Field(
        default=None,
        title="Callee",
        description="The address of the contract being called by this embedded calldata, as a constant value. Exactly "
        'one of "calleePath" or "callee" must be set.',
    )

    @model_validator(mode="after")
    def _validate_one_of_callee_path_or_value(self) -> Self:
        if self.calleePath is None and self.callee is None:
            raise ValueError('Either "calleePath" or "callee" must be set.')
        if self.calleePath is not None and self.callee is not None:
            raise ValueError('"calleePath" and "callee" are mutually exclusive.')
        return self


class InputNftNameParameters(Model):
    """
    NFT Names Formatting Parameters.
    """

    collectionPath: DescriptorPathStr | DataPathStr | ContainerPathStr | None = Field(
        default=None,
        title="Collection Path",
        description="The path to the collection in the structured data. Exactly one of "
        '"collectionPath" or "collection" must be set.',
    )

    collection: DescriptorPathStr | MixedCaseAddress | None = Field(
        default=None,
        title="Collection",
        description="The address of the collection contract, as a constant value. Exactly one of "
        '"collectionPath" or "collection" must be set.',
    )

    @model_validator(mode="after")
    def _validate_one_of_collection_path_or_value(self) -> Self:
        if self.collectionPath is None and self.collection is None:
            raise ValueError('Either "collectionPath" or "collection" must be set.')
        if self.collectionPath is not None and self.collection is not None:
            raise ValueError('"collectionPath" and "collection" are mutually exclusive.')
        return self


class InputDateParameters(Model):
    """
    Date Formatting Parameters
    """

    encoding: DateEncoding | DescriptorPathStr = Field(title="Date Encoding", description="The encoding of the date.")


class InputUnitParameters(Model):
    """
    Unit Formatting Parameters.
    """

    base: DescriptorPathStr | str = Field(
        title="Unit base symbol",
        description="The base symbol of the unit, displayed after the converted value. It can be an SI unit symbol or "
        "acceptable dimensionless symbols like % or bps.",
    )

    decimals: int | DescriptorPathStr | None = Field(
        default=None, title="Decimals", description="The number of decimals of the value, used to convert to a float."
    )

    prefix: bool | DescriptorPathStr | None = Field(
        default=None,
        title="Prefix",
        description="Whether the value should be converted to a prefixed unit, like k, M, G, etc.",
    )


class InputEnumParameters(Model):
    """
    Enum Formatting Parameters.
    """

    ref: DescriptorPathStr = Field(
        alias="$ref",
        title="Enum reference",
        description="The internal path to the enum definition used to convert this value.",
    )


InputFieldParameters = Annotated[
    Annotated[InputAddressNameParameters, Tag("address_name")]
    | Annotated[InputCallDataParameters, Tag("call_data")]
    | Annotated[InputTokenAmountParameters, Tag("token_amount")]
    | Annotated[InputNftNameParameters, Tag("nft_name")]
    | Annotated[InputDateParameters, Tag("date")]
    | Annotated[InputUnitParameters, Tag("unit")]
    | Annotated[InputEnumParameters, Tag("enum")],
    Discriminator(field_parameters_discriminator),
]


class InputFieldDefinition(Model):
    """
    A field formatter, containing formatting information of a single field in a message.
    """

    id: Id | None = Field(
        alias="$id",
        default=None,
        title="Id",
        description="An internal identifier that can be used either for clarity specifying what the element is or as a "
        "reference in device specific sections.",
    )

    label: DescriptorPathStr | str = Field(
        title="Field Label",
        description="The label of the field, that will be displayed to the user in front of the formatted field value.",
    )

    format: FieldFormat | None = Field(
        default=None,
        title="Field Format",
        description="The format of the field, that will be used to format the field value in a human readable way.",
    )

    params: InputFieldParameters | None = Field(
        default=None,
        title="Format Parameters",
        description="Format specific parameters that are used to format the field value in a human readable way.",
    )


class InputFieldDescription(InputFieldBase, InputFieldDefinition):
    """
    A field formatter, containing formatting information of a single field in a message.
    """


class InputNestedFields(InputFieldBase):
    """
    A single set of field formats, allowing recursivity in the schema.

    Used to group whole definitions for structures for instance. This allows nesting definitions of formats, but note
    that support for deep nesting will be device dependent.
    """

    fields: list[ForwardRef("InputField")] = Field(  # type: ignore
        title="Fields", description="Nested fields formats."
    )


InputField = Annotated[
    Annotated[InputReference, Tag("reference")]
    | Annotated[InputFieldDescription, Tag("field_description")]
    | Annotated[InputNestedFields, Tag("nested_fields")],
    Discriminator(field_discriminator),
]


class InputFormat(FormatBase):
    """
    A structured data format specification, containing formatting information of fields in a single type of message.
    """

    fields: list[InputField] = Field(
        title="Field Formats set",
        description="An array containing the ordered definitions of fields formats.",
        min_length=1,
    )

    required: list[DataPathStr | ContainerPathStr] | None = Field(
        default=None,
        title="Required fields",
        description="A list of fields that are required to be displayed to the user. A field that has a formatter and "
        "is not in this list is optional. A field that does not have a formatter should be silent, ie not "
        "shown.",
    )

    excluded: list[DataPathStr] | None = Field(
        default=None,
        title="Excluded fields",
        description="Intentionally excluded fields, as an array of *paths* referring to specific fields. A field that "
        "has no formatter and is not declared in this list MAY be considered as an error by the wallet when "
        "interpreting the descriptor. The excluded paths should interpreted as prefixes, meaning that all fields under "
        "excluded path should be ignored",
    )


class InputDisplay(Model):
    """
    Display Formatting Info Section.
    """

    definitions: dict[str, InputFieldDefinition] | None = Field(
        default=None,
        title="Common Formatter Definitions",
        description="A set of definitions that can be used to share formatting information between multiple messages / "
        "functions. The definitions can be referenced by the key name in an internal path.",
    )

    formats: dict[str, InputFormat] = Field(
        title="List of field formats",
        description="The list includes formatting info for each field of a structure. This list is indexed by a key "
        "identifying uniquely the message's type in the abi. For smartcontracts, it is the selector of the "
        "function or its signature; and for EIP712 messages it is the primaryType of the message.",
    )



from typing import assert_never, cast

from erc7730.common.abi import ABIDataType
from erc7730.common.output import OutputAdder
from erc7730.convert.resolved.constants import ConstantProvider
from erc7730.convert.resolved.enums import get_enum, get_enum_id
from erc7730.convert.resolved.values import resolve_path_or_constant_value
from erc7730.model.input.display import (
    InputAddressNameParameters,
    InputCallDataParameters,
    InputDateParameters,
    InputEnumParameters,
    InputFieldParameters,
    InputNftNameParameters,
    InputTokenAmountParameters,
    InputUnitParameters,
)
from erc7730.model.input.path import DescriptorPathStr
from erc7730.model.metadata import EnumDefinition
from erc7730.model.paths import DataPath
from erc7730.model.resolved.display import (
    ResolvedAddressNameParameters,
    ResolvedCallDataParameters,
    ResolvedDateParameters,
    ResolvedEnumParameters,
    ResolvedFieldParameters,
    ResolvedNftNameParameters,
    ResolvedTokenAmountParameters,
    ResolvedUnitParameters,
)
from erc7730.model.types import Address, HexStr, Id, MixedCaseAddress


def resolve_field_parameters(
    prefix: DataPath,
    params: InputFieldParameters | None,
    enums: dict[Id, EnumDefinition],
    constants: ConstantProvider,
    out: OutputAdder,
) -> ResolvedFieldParameters | None:
    match params:
        case None:
            return None
        case InputAddressNameParameters():
            return resolve_address_name_parameters(prefix, params, constants, out)
        case InputCallDataParameters():
            return resolve_calldata_parameters(prefix, params, constants, out)
        case InputTokenAmountParameters():
            return resolve_token_amount_parameters(prefix, params, constants, out)
        case InputNftNameParameters():
            return resolve_nft_parameters(prefix, params, constants, out)
        case InputDateParameters():
            return resolve_date_parameters(prefix, params, constants, out)
        case InputUnitParameters():
            return resolve_unit_parameters(prefix, params, constants, out)
        case InputEnumParameters():
            return resolve_enum_parameters(prefix, params, enums, constants, out)
        case _:
            assert_never(params)


def resolve_address_name_parameters(
    prefix: DataPath, params: InputAddressNameParameters, constants: ConstantProvider, out: OutputAdder
) -> ResolvedAddressNameParameters | None:
    return ResolvedAddressNameParameters(
        types=constants.resolve_or_none(params.types, out), sources=constants.resolve_or_none(params.sources, out)
    )


def resolve_calldata_parameters(
    prefix: DataPath, params: InputCallDataParameters, constants: ConstantProvider, out: OutputAdder
) -> ResolvedCallDataParameters | None:
    if (
        callee := resolve_path_or_constant_value(
            prefix=prefix,
            input_path=params.calleePath,
            input_value=params.callee,
            abi_type=ABIDataType.ADDRESS,
            constants=constants,
            out=out,
        )
    ) is None:
        return None

    return ResolvedCallDataParameters(
        selector=constants.resolve_or_none(params.selector, out),
        callee=callee,
    )


def resolve_token_amount_parameters(
    prefix: DataPath, params: InputTokenAmountParameters, constants: ConstantProvider, out: OutputAdder
) -> ResolvedTokenAmountParameters | None:
    token = resolve_path_or_constant_value(
        prefix=prefix,
        input_path=params.tokenPath,
        input_value=params.token,
        abi_type=ABIDataType.ADDRESS,
        constants=constants,
        out=out,
    )

    input_addresses = cast(
        list[DescriptorPathStr | MixedCaseAddress] | MixedCaseAddress | None,
        constants.resolve_or_none(params.nativeCurrencyAddress, out),
    )
    resolved_addresses: list[Address] | None
    if input_addresses is None:
        resolved_addresses = None
    elif isinstance(input_addresses, list):
        resolved_addresses = []
        for input_address in input_addresses:
            if (resolved_address := constants.resolve(input_address, out)) is None:
                return None
            resolved_addresses.append(Address(resolved_address))
    elif isinstance(input_addresses, str):
        resolved_addresses = [Address(input_addresses)]
    else:
        raise Exception("Invalid nativeCurrencyAddress type")

    input_threshold = cast(HexStr | int | None, constants.resolve_or_none(params.threshold, out))
    resolved_threshold: HexStr | None
    if input_threshold is not None:
        if isinstance(input_threshold, int):
            resolved_threshold = "0x" + input_threshold.to_bytes(byteorder="big", signed=False).hex()
        else:
            resolved_threshold = input_threshold
    else:
        resolved_threshold = None

    return ResolvedTokenAmountParameters(
        token=token,
        nativeCurrencyAddress=resolved_addresses,
        threshold=resolved_threshold,
        message=constants.resolve_or_none(params.message, out),
    )


def resolve_nft_parameters(
    prefix: DataPath, params: InputNftNameParameters, constants: ConstantProvider, out: OutputAdder
) -> ResolvedNftNameParameters | None:
    if (
        collection := resolve_path_or_constant_value(
            prefix=prefix,
            input_path=params.collectionPath,
            input_value=params.collection,
            abi_type=ABIDataType.ADDRESS,
            constants=constants,
            out=out,
        )
    ) is None:
        return None

    return ResolvedNftNameParameters(collection=collection)


def resolve_date_parameters(
    prefix: DataPath, params: InputDateParameters, constants: ConstantProvider, out: OutputAdder
) -> ResolvedDateParameters | None:
    return ResolvedDateParameters(encoding=constants.resolve(params.encoding, out))


def resolve_unit_parameters(
    prefix: DataPath, params: InputUnitParameters, constants: ConstantProvider, out: OutputAdder
) -> ResolvedUnitParameters | None:
    return ResolvedUnitParameters(
        base=constants.resolve(params.base, out),
        decimals=constants.resolve_or_none(params.decimals, out),
        prefix=constants.resolve_or_none(params.prefix, out),
    )


def resolve_enum_parameters(
    prefix: DataPath,
    params: InputEnumParameters,
    enums: dict[Id, EnumDefinition],
    constants: ConstantProvider,
    out: OutputAdder,
) -> ResolvedEnumParameters | None:
    if (enum_id := get_enum_id(params.ref, out)) is None:
        return None
    if get_enum(params.ref, enums, out) is None:
        return None

    return ResolvedEnumParameters(enumId=enum_id)



from typing import Any


def has_any_property(target: Any, *names: str) -> bool:
    """
    Check if the target has a property with any of the given names.

    :param target: object of dict like
    :param names: attribute names
    :return: true if the target has the property
    """
    return any(has_property(target, n) for n in names)


def has_property(target: Any, name: str) -> bool:
    """
    Check if the target has a property with the given name.

    :param target: object of dict like
    :param name: attribute name
    :return: true if the target has the property
    """
    if isinstance(target, dict):
        return name in target
    return hasattr(target, name)


def get_property(target: Any, name: str) -> Any:
    """
    Get the property with the given name on target object.

    :param target: object of dict like
    :param name: attribute name
    :return: value for property on target object
    """
    if isinstance(target, dict):
        return target[name]
    return getattr(target, name)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, assert_never, override

from pydantic import TypeAdapter, ValidationError
from typing_extensions import TypeVar

from erc7730.common.output import OutputAdder
from erc7730.common.properties import get_property
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.input.path import ContainerPathStr, DataPathStr
from erc7730.model.paths import ROOT_DESCRIPTOR_PATH, ArrayElement, ContainerPath, DataPath, DescriptorPath, Field
from erc7730.model.paths.path_ops import descriptor_path_append, to_absolute
from erc7730.model.types import MixedCaseAddress

_T = TypeVar("_T", covariant=True)


class ConstantProvider(ABC):
    """
    Resolver for constants values referenced by descriptor paths.
    """

    @abstractmethod
    def get(self, path: DescriptorPath, out: OutputAdder) -> Any:
        """
        Get the constant for the given path.

        :param path: descriptor path
        :param out: error handler
        :return: constant value, or None if not found
        """
        raise NotImplementedError()

    def resolve(self, value: _T | DescriptorPath, out: OutputAdder) -> _T:
        """
        Resolve the value if it is a descriptor path.

        :param value: descriptor path or actual value
        :param out: error handler
        :return: constant value, or the value itself if not a descriptor path
        """
        return self.get(value, out) if isinstance(value, DescriptorPath) else value

    def resolve_or_none(self, value: _T | DescriptorPath | None, out: OutputAdder) -> _T | None:
        """
        Resolve the optional value if it is a descriptor path.

        :param value: descriptor path, actual value or None
        :param out: error handler
        :return: None, constant value, or the value itself if not a descriptor path
        """
        return None if value is None else self.resolve(value, out)

    def resolve_path(
        self, value: DataPath | ContainerPath | DescriptorPath, out: OutputAdder
    ) -> DataPath | ContainerPath | None:
        """
        Resolve the value as a data/container path.

        :param value: descriptor path or actual data/container path
        :param out: error handler
        :return: resolved data/container path
        """

        def assert_not_address(path: DataPath | ContainerPath) -> bool:
            match path:
                case ContainerPath():
                    return True
                case DataPath():
                    if path.absolute:
                        return True
                    try:
                        TypeAdapter(MixedCaseAddress).validate_strings(str(path))
                        out.error(
                            title="Invalid data path",
                            message=f""""{path}" is invalid, it must contain a data path to the address in the """
                            "transaction data. It seems you are trying to use a constant address value instead, please "
                            "use the adequate parameter to provide a constant value.",
                        )
                        return False
                    except ValidationError:
                        return True
                case _:
                    assert_never(path)

        if isinstance(value, DataPath | ContainerPath):
            if not assert_not_address(value):
                return None
            return value

        resolved_value: Any
        if (resolved_value := self.resolve(value, out)) is None:
            return None

        if not isinstance(resolved_value, str):
            return out.error(
                title="Invalid constant path",
                message=f"Constant path defined at {value} must be a path string, got {type(resolved_value).__name__}.",
            )

        match TypeAdapter(DataPathStr | ContainerPathStr).validate_strings(resolved_value):
            case ContainerPath() as path:
                return path
            case DataPath() as path:
                if not assert_not_address(path):
                    return None
                if not path.absolute:
                    return out.error(
                        title="Invalid data path constant",
                        message=f"Data path defined at {value} must be absolute, please change it to "
                        f"{to_absolute(path)}.",
                    )
                return path
            case _:
                assert_never(resolved_value)

        # noinspection PyUnreachableCode
        return None

    def resolve_path_or_none(
        self, value: DataPath | ContainerPath | DescriptorPath | None, out: OutputAdder
    ) -> DataPath | ContainerPath | None:
        """
        Resolve the value as a data/container path.

        :param value: descriptor path or actual data/container path
        :param out: error handler
        :return: resolved data/container path
        """
        return None if value is None else self.resolve_path(value, out)


class DefaultConstantProvider(ConstantProvider):
    """
    Resolver for constants values from a provided dictionary.
    """

    def __init__(self, descriptor: InputERC7730Descriptor) -> None:
        self.descriptor: InputERC7730Descriptor = descriptor

    @override
    def get(self, path: DescriptorPath, out: OutputAdder) -> Any:
        current_target = self.descriptor
        parent_path = ROOT_DESCRIPTOR_PATH
        current_path = ROOT_DESCRIPTOR_PATH

        for element in path.elements:
            current_path = descriptor_path_append(current_path, element)
            match element:
                case Field(identifier=field):
                    if isinstance(current_target, Sequence):
                        return out.error(
                            title="Invalid constant path",
                            message=f"""Path {current_path} is invalid, {parent_path} is an array.""",
                        )
                    else:
                        try:
                            current_target = get_property(current_target, field)
                        except (AttributeError, KeyError):
                            return out.error(
                                title="Invalid constant path",
                                message=f"""Path {current_path} is invalid, {parent_path} has no "{field}" field.""",
                            )
                case ArrayElement(index=i):
                    if not isinstance(current_target, Sequence):
                        return out.error(
                            title="Invalid constant path",
                            message=f"Path {current_path} is invalid, {parent_path} is not an array.",
                        )
                    if i >= len(current_target):
                        return out.error(
                            title="Invalid constant path",
                            message=f"""Path {current_path} is invalid, index {i} is out of bounds.""",
                        )
                    current_target = current_target[i]
                case _:
                    assert_never(element)
            parent_path = descriptor_path_append(parent_path, element)

        return current_target



from typing_extensions import TypeVar

_T = TypeVar("_T")


def first_not_none(*args: _T | None) -> _T | None:
    """
    Return the first argument that is not None.

    :param args: sequence of optional values
    :return: first non-None value, or None if there are none
    """
    for arg in args:
        if arg is not None:
            return arg
    return None



from typing import final, override

from erc7730.common import client
from erc7730.common.abi import compute_signature, get_functions
from erc7730.common.output import OutputAdder
from erc7730.lint import ERC7730Linter
from erc7730.model.resolved.context import ResolvedContractContext, ResolvedEIP712Context
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor


@final
class ValidateABILinter(ERC7730Linter):
    """
    - resolves the ABI from the descriptor (URL or provided)
    - resolves the ABI from *scan (given chainId and address of descriptor)
    - => compares the two ABIs
    """

    @override
    def lint(self, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        if isinstance(descriptor.context, ResolvedEIP712Context):
            return self._validate_eip712_schemas(descriptor.context, out)
        if isinstance(descriptor.context, ResolvedContractContext):
            return self._validate_contract_abis(descriptor.context, out)
        raise ValueError("Invalid context type")

    @classmethod
    def _validate_eip712_schemas(cls, context: ResolvedEIP712Context, out: OutputAdder) -> None:
        pass  # not implemented

    @classmethod
    def _validate_contract_abis(cls, context: ResolvedContractContext, out: OutputAdder) -> None:
        if not isinstance(context.contract.abi, list):
            raise ValueError("Contract ABIs should have been resolved")

        if (deployments := context.contract.deployments) is None:
            return
        for deployment in deployments:
            try:
                if (abis := client.get_contract_abis(deployment.chainId, deployment.address)) is None:
                    continue
            except Exception as e:
                out.warning(
                    title="Could not fetch ABI",
                    message=f"Fetching reference ABI for chain id {deployment.chainId} failed, descriptor ABIs will "
                    f"not be validated: {e}",
                )
                continue

            reference_abis = get_functions(abis)
            descriptor_abis = get_functions(context.contract.abi)
            url = client.get_contract_explorer_url(deployment.chainId, deployment.address)

            if reference_abis.proxy:
                return out.info(
                    title="Proxy contract",
                    message=f"Contract {url} is likely to be a proxy, validation of descriptor ABIs skipped",
                )

            for selector, abi in descriptor_abis.functions.items():
                if selector not in reference_abis.functions:
                    out.warning(
                        title="Extra function",
                        message=f"Function {compute_signature(abi)} (selector: {selector}) defined in descriptor ABIs "
                        f"does not exist in reference ABI (see {url})",
                    )
                elif descriptor_abis.functions[selector] != reference_abis.functions[selector]:
                    out.warning(
                        title="Function mismatch",
                        message=f"Function {compute_signature(abi)} (selector: {selector}) defined in descriptor ABIs "
                        f"does not match reference ABI (see {url})",
                    )



from typing import Annotated

from pydantic import AfterValidator, Field

from erc7730.model.paths import ContainerPath, DataPath


def _validate_absolute(path: ContainerPath | DataPath) -> ContainerPath | DataPath:
    if isinstance(path, ContainerPath):
        return path
    if isinstance(path, DataPath):
        if not path.absolute:
            raise ValueError("A resolved data path must be absolute")
        return path
    raise ValueError(f"Invalid path type: {type(path)}")


ResolvedPath = Annotated[
    ContainerPath | DataPath,
    Field(
        title="Resolved Path",
        description="A path in the input designating value(s) either in the container of the structured data to be"
        "signed or the structured data schema (ABI path for contracts, path in the message types itself for EIP-712).",
        discriminator="type",
    ),
    AfterValidator(_validate_absolute),
]



from collections.abc import Generator
from pathlib import Path

from rich import print

from erc7730 import ERC_7730_REGISTRY_CALLDATA_PREFIX, ERC_7730_REGISTRY_EIP712_PREFIX
from erc7730.common.output import (
    ConsoleOutputAdder,
    OutputAdder,
)


def list_all(paths: list[Path]) -> bool:
    """
    List all ERC-7730 descriptor files at given paths.

    Paths can be files or directories, in which case all descriptor files in the directory are recursively listed.

    :param paths: paths to search for descriptor files
    :return: true if no error occurred
    """
    out = ConsoleOutputAdder()

    for file in get_erc7730_files(*paths, out=out):
        print(file)

    return not out.has_errors


def get_erc7730_files(*paths: Path, out: OutputAdder) -> Generator[Path, None, None]:
    """
    List all ERC-7730 descriptor files at given paths.

    Paths can be files or directories, in which case all descriptor files in the directory are recursively listed.

    :param paths: paths to search for descriptor files
    :param out: error handler
    """
    for path in paths:
        if path.is_file():
            if is_erc7730_file(path):
                yield path
            else:
                out.error(title="Invalid path", message=f"{path} is not an ERC-7730 descriptor file")
        elif path.is_dir():
            for file in path.rglob("*.json"):
                if is_erc7730_file(file):
                    yield file
        else:
            out.error(title="Invalid path", message=f"{path} is not a file or directory")


def is_erc7730_file(path: Path) -> bool:
    """
    Check if a file is an ERC-7730 descriptor file.

    :param path: file path
    :return: true if the file is an ERC-7730 descriptor file
    """
    return path.is_file() and (
        path.name.startswith(ERC_7730_REGISTRY_CALLDATA_PREFIX) or path.name.startswith(ERC_7730_REGISTRY_EIP712_PREFIX)
    )



from typing import Any

from erc7730.common.properties import has_any_property


def field_discriminator(v: Any) -> str | None:
    """
    Discriminator function for the Field union type.

    :param v: deserialized raw data
    :return: the discriminator tag
    """
    if has_any_property(v, "$ref"):
        return "reference"
    if has_any_property(v, "fields"):
        return "nested_fields"
    if has_any_property(v, "label"):
        return "field_description"
    return None


def field_parameters_discriminator(v: Any) -> str | None:
    """
    Discriminator function for the FieldParameters union type.

    :param v: deserialized raw data
    :return: the discriminator tag
    """
    if has_any_property(v, "tokenPath", "token", "nativeCurrencyAddress"):
        return "token_amount"
    if has_any_property(v, "encoding"):
        return "date"
    if has_any_property(v, "collectionPath", "collection"):
        return "nft_name"
    if has_any_property(v, "base"):
        return "unit"
    if has_any_property(v, "$ref", "ref", "enumId"):
        return "enum"
    if has_any_property(v, "calleePath", "callee", "selector"):
        return "call_data"
    if has_any_property(v, "sources", "types"):
        return "address_name"
    return None



#!/usr/bin/env python3
import glob
import hashlib
import json
import logging
import re
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Callable, Tuple, List

from eth_utils.address import to_checksum_address
from jsonschema import ValidationError
from jsonschema.validators import validator_for

logger = logging.getLogger(__name__)


def run_validations(glob_path: str, validator: Callable[[str, str], Tuple[bool, str]]):
    valid, invalid, errors = 0, 0, []
    for filename in glob.iglob(glob_path, recursive=True):
        logger.debug("Validating %s...", filename)
        with open(filename, "r", newline="") as f:
            data = f.read()
        is_valid, error_message = validator(data, filename)
        if is_valid:
            valid += 1
        else:
            invalid += 1
            errors.append({"file": filename, "message": error_message})

    logger.info(f"Result: {valid} valid, {invalid} invalid")
    if invalid:
        for error in errors:
            logger.error(f"Validation error: {error}")
        raise ValidationError(f"Invalid files: {errors}")


# Some DApps using these contracts are already signed and have collisions.
# Until we know how to correctly handle these cases
# we exclude them from the collision check
# https://ledgerhq.atlassian.net/wiki/spaces/BE/pages/4623073281/DApps+contract+collision+analysis
__excluded_contract = [
    "0xdef171fe48cf0115b1d80b88dc8eab59176fee57",
    "0x1111111254fb6c44bac0bed2854e76f90643097d",
    "0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2",
]

__excluded_contracts_from_plugin_collision = [
    "0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2",
]


def check_duplicate_contract(glob_path: str):
    invalid, errors, results, all = 0, [], [], {}
    for filename in glob.iglob(glob_path, recursive=True):
        logger.debug("Validating %s...", filename)
        with open(filename, "r", newline="") as f:
            data = f.read()

        try:
            target_path = Path(filename)
            target_data = json.loads(data)
        except JSONDecodeError as err:
            logger.debug(
                "\tinvalid: File %s is not a valid json", filename, exc_info=True
            )
            errors.append({"file": filename, "message": str(err)})
            continue

        result = {}
        blockChain = target_data.get("blockchainName", None)
        name = target_data.get("name", None)
        for contract in target_data.get("contracts", []):
            address = contract.get("address", None)
            if address in __excluded_contract:
                continue
            elif address is None:
                continue
            address = address.lower()

            abi_path = target_path.parent / "abis" / f"{address}.abi.json"
            with open(abi_path, "r", newline="") as f:
                try:
                    abi_data = json.load(f)
                except JSONDecodeError as err:
                    logger.debug(
                        "\tinvalid: File %s is not a valid json",
                        abi_path,
                        exc_info=True,
                    )
                    errors.append({"file": abi_path, "message": str(err)})
                    continue

            for data in abi_data:
                for input in data.get("inputs", []):
                    input.pop("name", None)

            selectors = {}
            for selector, data in contract.get("selectors", []).items():
                selectors[selector] = data["plugin"]

            result[address] = {
                "blockChain": blockChain,
                "name": name,
                "abi": abi_data,
                "filename": filename,
                "selectors": {**selectors},
            }

            results.append(result)

    for result in results:
        for contract_address, data in result.items():
            invalid_inter, error = 0, []
            existing = all.get(contract_address, None)
            if existing is None:
                all[contract_address] = data
            else:
                if existing is not None:
                    if existing["abi"] != data["abi"]:
                        invalid_inter += 1
                        error.append("seem to have two different abi")
                    if existing["selectors"] != data["selectors"]:
                        invalid_inter += 1
                        error.append("seem to have differents selectors")
                    if invalid_inter:
                        invalid += 1
                        errors.append(
                            {
                                "file": data["filename"],
                                "message": f"plugin contract {contract_address} "
                                + " and ".join(error),
                            }
                        )
                    all[contract_address] = {**existing, **data}

    if invalid:
        for error in errors:
            logger.error(f"Validation error: {error}")
        raise ValidationError(f"Invalid files: {errors}")


def check_duplicate_plugin(glob_path: str):
    invalid, errors, results, all = 0, [], [], {}
    for filename in glob.iglob(glob_path, recursive=True):
        logger.debug("Validating %s... for duplicate plugins", filename)
        with open(filename, "r", newline="") as f:
            data = f.read()

        try:
            target_data = json.loads(data)
        except JSONDecodeError as err:
            logger.debug(
                "\tinvalid: File %s is not a valid json", filename, exc_info=True
            )
            errors.append({"file": filename, "message": str(err)})
            continue

        records = []
        blockchain = target_data.get("blockchainName", None)
        for contract in target_data.get("contracts", []):
            address = contract.get("address", None)
            if address is None:
                continue
            elif address in __excluded_contracts_from_plugin_collision:
                continue

            for selector, data in contract.get("selectors", []).items():
                plugin = data["plugin"]

                record = {
                    "filename": filename,
                    "blockchain": blockchain,
                    "contract": address.lower(),
                    "selector": selector.lower(),
                    "plugin": plugin,
                }
                logger.debug(f"adding record {record}")
                records.append(record)

        # check if among the records there are multiple plugins associated with a given (chain_id, contract, selector) triplet
        for record in records:
            filename = record["filename"]
            blockchain = record["blockchain"]
            contract = record["contract"]
            selector = record["selector"]
            plugin = record["plugin"]
            existing_record = all.get((blockchain, contract, selector), None)
            if existing_record is None:
                all[(blockchain, contract, selector)] = record
            elif existing_record["plugin"] != plugin:
                invalid += 1
                errors.append(f"File {filename} and file {existing_record['filename']} bind "
                              f"(blockchain={blockchain}, contract={contract}, selector={selector}) to different plugins "
                              f"{plugin} and {existing_record['plugin']} respectively, "
                              f"please remove one of the bindings")

    if invalid:
        for error in errors:
            logger.error(f"Validation error: {error}")
        raise ValidationError(f"Invalid files: {errors}")


def format_validator(data: str, filename: str) -> Tuple[bool, str]:
    try:
        loaded = json.loads(data)
        formatted = json.dumps(loaded, indent=4, sort_keys=True, ensure_ascii=False)
        if formatted != data and (formatted + "\n") != data:
            logger.debug(
                "\tinvalid: File %s is not formatted correctly ; expected: %s",
                filename,
                formatted,
            )
            with open(filename, "w") as f:
                json.dump(loaded, f, indent=4, sort_keys=True, ensure_ascii=False)
            return False, "Not formatted correctly. Fixed."
    except JSONDecodeError as err:
        logger.debug("\tinvalid: File %s is not a valid json", filename, exc_info=True)
        return False, str(err)
    return True, ""


def endlines_validator(data: str, filename: str) -> Tuple[bool, str]:
    is_valid = "\\r\\n" not in data
    error_message = ""
    if not is_valid:
        error_message = "File has some endlines chars"
        logger.debug("\tinvalid: File %s has some endlines chars", filename)
    return is_valid, error_message


def unique_field_validator(field_names: List[str]):
    unique = dict()

    def _inner_validator(data: str, filename: str) -> Tuple[bool, str]:
        try:
            json_data = json.loads(data)
        except JSONDecodeError as err:
            logger.debug(
                "\tinvalid: File %s is not a valid json", filename, exc_info=True
            )
            return False, str(err)
        field_value = next(
            json_data[field_name]
            for field_name in field_names
            if field_name in json_data
        )
        coin_name = Path(filename).parts[0]
        identifier = f"{coin_name}/{field_value}".lower()

        if identifier in unique:
            logger.info(
                "\tinvalid: File %s has the same %s as %s",
                filename,
                field_names,
                unique[identifier],
            )
            return False, f"{field_names} is not unique."
        else:
            unique[identifier] = filename
            return True, ""

    return _inner_validator


def schema_validator(schema_path: str):
    with open(schema_path, "r") as f:
        schema = json.load(f)
    validator_class = validator_for(schema)
    validator_class.check_schema(schema)
    validator = validator_class(schema)

    def _inner_validator(data: str, filename: str) -> Tuple[bool, str]:
        try:
            validator.validate(json.loads(data))
            return True, ""
        except ValidationError as err:
            logger.debug(
                "\tinvalid: File %s doesn't match the schema", filename, exc_info=True
            )
            return False, err.message

    return _inner_validator


def eip712_schema_validator(data: str, filename: str) -> Tuple[bool, str]:
    try:
        loaded = json.loads(data)
        for contract in loaded["contracts"]:
            for message in contract["messages"]:
                # schema types must be sorted by key for reproducible hashing
                message["schema"] = dict(sorted(message["schema"].items()))

        formatted = json.dumps(loaded, indent=4, sort_keys=True, ensure_ascii=False)
        if formatted != data and (formatted + "\n") != data:
            logger.debug(
                "\tinvalid: File %s is not formatted correctly ; expected: %s",
                filename,
                formatted,
            )
            with open(filename, "w") as f:
                f.write(formatted)
            return False, "Not formatted, corrected"
    except JSONDecodeError as err:
        logger.debug("\tinvalid: File %s is not a valid json", filename, exc_info=True)
        return False, str(err)
    return True, ""


def missing_schema_validator():
    coins = set()

    def _inner_validator(data: str, filename: str) -> Tuple[bool, str]:
        parser_path = Path(filename)
        coin_name = parser_path.parts[0]
        if coin_name in coins:
            return True, ""
        coins.add(coin_name)
        target_name = parser_path.name.removesuffix(".json")
        schema_path = parser_path.parent.parent / f"{target_name}.schema.json"
        if schema_path.exists():
            return True, ""
        else:
            return False, f"Missing schema {schema_path}"

    return _inner_validator


def missing_abi_validator(data: str, filename: str) -> Tuple[bool, str]:
    try:
        dapp_addresses = set(
            contract.get("address", "").lower()
            for contract in json.loads(data).get("contracts", [])
        )
        abi_addresses = set(
            p.name.split(".")[0].lower()
            for p in Path(filename).parent.glob("abis/*.abi.json")
        )
        if not dapp_addresses.issubset(abi_addresses):
            return (
                False,
                f"Missing ABI for contract {dapp_addresses.difference(abi_addresses)}",
            )
    except Exception as err:
        return False, str(err)
    return True, ""


def abi_filename_validator(data: str, filename: str) -> Tuple[bool, str]:
    lowercase_address_regex = r"^0x[a-f0-9]{40}\.abi\.json$"
    if re.match(lowercase_address_regex, Path(filename).name):
        return True, ""
    else:
        return False, f"ABI filename is not matching {lowercase_address_regex}"


def eip55_address_validator(data: str, filename: str) -> Tuple[bool, str]:
    error = False
    try:
        json_data = json.loads(data)
    except JSONDecodeError as err:
        logger.debug("\tinvalid: File %s is not a valid json", filename, exc_info=True)
        return False, str(err)

    for contract in json_data.get("contracts", []):
        # Address respects EIP55 format
        address = contract.get("address", "")
        expected = to_checksum_address(address)
        is_valid = address == expected
        if not is_valid:
            error = True
            logger.info(
                "\tinvalid: Contract address %s not in eip55 format",
                address,
            )
            contract["address"] = expected
    if error:
        formatted = json.dumps(json_data, indent=4, sort_keys=True, ensure_ascii=False)
        with open(filename, "w") as f:
            f.write(formatted)
        return False, "Some addresses not correctly formatted, fixing..."

    return True, ""


def contract_matching_validator(data: str, filename: str) -> Tuple[bool, str]:
    error = False
    try:
        target_path = Path(filename)
        target_data = json.loads(data)
    except JSONDecodeError as err:
        logger.debug("\tinvalid: File %s is not a valid json", filename, exc_info=True)
        return False, str(err)

    for contract in target_data.get("contracts", []):
        logger.debug(
            "\tchecking contract %s...",
            contract["address"],
        )
        abi_path = (
            target_path.parent / "abis" / f"{contract['address'].lower()}.abi.json"
        )
        abi_data = json.load(abi_path.open())

        target_functions = {
            parser["functionName"]: parser for parser in contract.get("parsers", [])
        }
        abi_functions = {
            function["name"]: function
            for function in abi_data
            if function["type"] == "function"
        }

        for name, function in target_functions.items():
            if name not in abi_functions:
                error = True
                logger.info(
                    "\tinvalid: Function %s not defined in ABI",
                    function,
                )
                continue

            logger.debug(
                "\tchecking function %s...",
                name,
            )

            function_args = {arg["name"] for arg in function.get("arguments", [])}
            screen_args = {
                arg["name"]
                for arg in function.get("screen", [])
                if arg["label"] != "Function"
            }
            if not screen_args <= function_args:
                error = True
                logger.info(
                    "\tinvalid: Screen %s not matching function arguments %s.",
                    function_args,
                    screen_args,
                )

    if error:
        return False, "Errors found on file content"

    return True, ""



from pathlib import Path

import pytest

from erc7730.convert.convert import convert_and_raise_errors
from erc7730.convert.resolved.convert_erc7730_input_to_resolved import ERC7730InputToResolved
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor
from tests.assertions import assert_model_json_equals
from tests.cases import TestCase, case_id, path_id
from tests.files import ERC7730_DESCRIPTORS
from tests.skip import single_or_skip

DATA = Path(__file__).resolve().parent / "data"
UPDATE_REFERENCES = False


@pytest.mark.parametrize("input_file", ERC7730_DESCRIPTORS, ids=path_id)
def test_registry_files(input_file: Path) -> None:
    """
    Test converting ERC-7730 registry files from input to resolved form.
    """
    convert_and_raise_errors(InputERC7730Descriptor.load(input_file), ERC7730InputToResolved())


@pytest.mark.parametrize(
    "testcase",
    [
        TestCase(
            id="minimal_eip712",
            label="minimal EIP-712 descriptor",
            description="most minimal possible EIP-712 file: all optional fields unset, resolved form is identical to "
            "input form",
        ),
        TestCase(
            id="minimal_contract",
            label="minimal contract descriptor",
            description="most minimal possible contract file: all optional fields unset, resolved form is identical to "
            "input form",
        ),
        TestCase(
            id="format_raw",
            label="field format - using raw format",
            description="using raw format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_address_name",
            label="field format - using address name format",
            description="using address name format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_calldata",
            label="field format - using calldata format",
            description="using calldata format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_amount",
            label="field format - using amount format",
            description="using amount format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_token_amount",
            label="field format - using token amount format",
            description="using token amount format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_nft_name",
            label="field format - using NFT name amount format",
            description="using NFT name amount format, with parameter variants, resolved form is identical to input "
            "form",
        ),
        TestCase(
            id="format_date",
            label="field format - using date format",
            description="using date format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_duration",
            label="field format - using duration format",
            description="using duration format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_unit",
            label="field format - using unit format",
            description="using unit format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_enum",
            label="field format - using enum format",
            description="using enum format, with parameter variants, resolved form is identical to input form",
        ),
        TestCase(
            id="format_raw_using_constants",
            label="field format - using raw format with references to constants",
            description="using raw format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_address_name_using_constants",
            label="field format - using address name format with references to constants",
            description="using address name format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_calldata_using_constants",
            label="field format - using calldata format with references to constants",
            description="using calldata format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_amount_using_constants",
            label="field format - using amount format with references to constants",
            description="using amount format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_token_amount_using_constants",
            label="field format - using token amount format with references to constants",
            description="using token amount format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_nft_name_using_constants",
            label="field format - using NFT name amount format with references to constants",
            description="using NFT name amount format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_date_using_constants",
            label="field format - using date format with references to constants",
            description="using date format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_duration_using_constants",
            label="field format - using duration format with references to constants",
            description="using duration format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_unit_using_constants",
            label="field format - using unit format with references to constants",
            description="using unit format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="format_enum_using_constants",
            label="field format - using enum format with references to constants",
            description="using enum format, with parameter variants and $.context/$.metadata constants",
        ),
        TestCase(
            id="definition_format_raw",
            label="display definition / reference - using raw format",
            description="most minimal possible use of a display definition + reference, using raw format",
        ),
        TestCase(
            id="definition_format_address_name",
            label="display definition / reference - using address name format",
            description="most minimal possible use of a display definition + reference, using address name format",
        ),
        TestCase(
            id="definition_format_calldata",
            label="display definition / reference - using calldata format",
            description="most minimal possible use of a display definition + reference, using calldata format",
        ),
        TestCase(
            id="definition_format_amount",
            label="display definition / reference - using amount format",
            description="most minimal possible use of a display definition + reference, using amount format",
        ),
        TestCase(
            id="definition_format_token_amount",
            label="display definition / reference - using token amount format",
            description="most minimal possible use of a display definition + reference, using token amount format",
        ),
        TestCase(
            id="definition_format_nft_name",
            label="display definition / reference - using NFT name format",
            description="most minimal possible use of a display definition + reference, using NFT name format",
        ),
        TestCase(
            id="definition_format_date",
            label="display definition / reference - using date format",
            description="most minimal possible use of a display definition + reference, using date format",
        ),
        TestCase(
            id="definition_format_duration",
            label="display definition / reference - using duration format",
            description="most minimal possible use of a display definition + reference, using duration format",
        ),
        TestCase(
            id="definition_format_unit",
            label="display definition / reference - using unit format",
            description="most minimal possible use of a display definition + reference, using unit format",
        ),
        TestCase(
            id="definition_format_enum",
            label="display definition / reference - using enum format",
            description="most minimal possible use of a display definition + reference, using enum format",
        ),
        TestCase(
            id="definition_override_label",
            label="display definition / reference - using label override",
            description="use of a display definition, with label overridden on the field",
        ),
        TestCase(
            id="definition_override_params",
            label="display definition / reference - using params override",
            description="use of a display definition, with parameters overridden on the field",
        ),
        TestCase(
            id="definition_using_constants",
            label="display definition / reference - using constants override",
            description="use of a display definition, using $.context/$.metadata constants",
        ),
        TestCase(
            id="definition_invalid_container_path",
            label="display definition / reference - using a container path",
            description="use of a reference to a display definition with a container path",
            error="descriptor path referencing a value in the descriptor document using jsonpath syntax, such as",
        ),
        TestCase(
            id="definition_invalid_data_path",
            label="display definition / reference - using a data path",
            description="use of a reference to a display definition with a data path",
            error="descriptor path referencing a value in the descriptor document using jsonpath syntax, such as",
        ),
        TestCase(
            id="definition_invalid_path_does_not_exist",
            label="display definition / reference - using a path that does not exist",
            description="use of a reference to a display definition that does not exist",
            error="""Display definition "does_not_exist" does not exist, valid ones are: test_definition.""",
        ),
        TestCase(
            id="definition_invalid_path_not_a_field",
            label="display definition / reference - using a path that is not a field",
            description="use of a reference path that is not a simple field reference",
            error="References to display field definitions are restricted to fields immediately under "
            "$.display.definitions",
        ),
        TestCase(
            id="definition_invalid_path_outside_display_definitions",
            label="display definition / reference - using a path outside of $.display.definitions",
            description="use of a reference to a display outside of allowed root path $.display.definitions",
            error="References to display field definitions are restricted to $.display.definitions",
        ),
        TestCase(
            id="definition_invalid_nested_path",
            label="display definition / reference - using a path nested under $.display.definitions",
            description="use of a valid reference to a display definition nested too deep under $.display.definitions",
            error="References to display field definitions are restricted to fields immediately under "
            "$.display.definitions",
        ),
        TestCase(
            id="definition_invalid_parameter_overrides",
            label="display definition / reference - using invalid parameter overrides",
            description="use of a valid reference to a valid display definition, with invalid overrides",
            error="Extra inputs are not permitted",
        ),
        TestCase(
            id="nested_fields_eip712_array",
            label="nested fields - array parameter (EIP-712 descriptor)",
            description="use of nested fields on an array parameter (EIP-712 descriptor)",
        ),
        TestCase(
            id="nested_fields_eip712_struct",
            label="nested fields - struct parameter (EIP-712 descriptor)",
            description="use of nested fields on a struct parameter (EIP-712 descriptor)",
        ),
        TestCase(
            id="unsupported_using_literal_values",
            label="unsupported: using literal value instead of data path",
            description="using a literal value where a data path is expected",
            error="It seems you are trying to use a constant address value instead",
        ),
        TestCase(
            id="unsupported_using_literal_values_as_constants",
            label="unsupported: using literal value instead of data path",
            description="using a literal value where a data path is expected, defined in constants section",
            error="It seems you are trying to use a constant address value instead",
        ),
        TestCase(
            id="literal_values", label="using literal values", description="using literal values anywhere possible"
        ),
    ],
    ids=case_id,
)
def test_by_reference(testcase: TestCase) -> None:
    """
    Test converting ERC-7730 descriptor files from input to resolved form, and compare against reference files.
    """
    input_descriptor_path = DATA / f"{testcase.id}_input.json"
    resolved_descriptor_path = DATA / f"{testcase.id}_resolved.json"
    if (expected_error := testcase.error) is not None:
        with pytest.raises(Exception) as exc_info:
            input_descriptor = InputERC7730Descriptor.load(input_descriptor_path)
            convert_and_raise_errors(input_descriptor, ERC7730InputToResolved())
        assert expected_error in str(exc_info.value)
    else:
        input_descriptor = InputERC7730Descriptor.load(input_descriptor_path)
        actual_descriptor: ResolvedERC7730Descriptor = single_or_skip(
            convert_and_raise_errors(input_descriptor, ERC7730InputToResolved())
        )
        if UPDATE_REFERENCES:
            actual_descriptor.save(resolved_descriptor_path)
            pytest.fail(f"Reference {resolved_descriptor_path} updated, please set UPDATE_REFERENCES back to False")
        else:
            expected_descriptor = ResolvedERC7730Descriptor.load(resolved_descriptor_path)
            assert_model_json_equals(expected_descriptor, actual_descriptor)



from typing import assert_never, final, override

from eip712.model.schema import EIP712Type
from pydantic_string_url import HttpUrl

from erc7730.common import client
from erc7730.common.abi import reduce_signature, signature_to_selector
from erc7730.common.output import ExceptionsToOutput, OutputAdder
from erc7730.convert import ERC7730Converter
from erc7730.convert.resolved.constants import ConstantProvider, DefaultConstantProvider
from erc7730.convert.resolved.parameters import resolve_field_parameters
from erc7730.convert.resolved.references import resolve_reference
from erc7730.convert.resolved.values import resolve_field_value
from erc7730.model.abi import ABI
from erc7730.model.context import EIP712Schema
from erc7730.model.display import (
    FieldFormat,
)
from erc7730.model.input.context import (
    InputContract,
    InputContractContext,
    InputDeployment,
    InputDomain,
    InputEIP712,
    InputEIP712Context,
    InputFactory,
)
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.input.display import (
    InputDisplay,
    InputField,
    InputFieldDefinition,
    InputFieldDescription,
    InputFormat,
    InputNestedFields,
    InputReference,
)
from erc7730.model.input.metadata import InputMetadata
from erc7730.model.metadata import EnumDefinition
from erc7730.model.paths import ROOT_DATA_PATH, Array, ArrayElement, ArraySlice, ContainerPath, DataPath, Field
from erc7730.model.paths.path_ops import data_path_concat
from erc7730.model.resolved.context import (
    ResolvedContract,
    ResolvedContractContext,
    ResolvedDeployment,
    ResolvedDomain,
    ResolvedEIP712,
    ResolvedEIP712Context,
    ResolvedFactory,
)
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor
from erc7730.model.resolved.display import (
    ResolvedDisplay,
    ResolvedField,
    ResolvedFieldDescription,
    ResolvedFormat,
    ResolvedNestedFields,
    ResolvedValuePath,
)
from erc7730.model.resolved.metadata import ResolvedMetadata
from erc7730.model.types import Address, Id, Selector


@final
class ERC7730InputToResolved(ERC7730Converter[InputERC7730Descriptor, ResolvedERC7730Descriptor]):
    """
    Converts ERC-7730 descriptor input to resolved form.

    After conversion, the descriptor is in resolved form:
     - URLs have been fetched
     - Contract addresses have been normalized to lowercase
     - References have been inlined
     - Constants have been inlined
     - Field definitions have been inlined
     - Nested fields have been flattened where possible
     - Selectors have been converted to 4 bytes form
    """

    @override
    def convert(self, descriptor: InputERC7730Descriptor, out: OutputAdder) -> ResolvedERC7730Descriptor | None:
        with ExceptionsToOutput(out):
            constants = DefaultConstantProvider(descriptor)

            if (context := self._resolve_context(descriptor.context, out)) is None:
                return None
            if (metadata := self._resolve_metadata(descriptor.metadata, out)) is None:
                return None
            if (display := self._resolve_display(descriptor.display, context, metadata.enums, constants, out)) is None:
                return None

            return ResolvedERC7730Descriptor(context=context, metadata=metadata, display=display)

        # noinspection PyUnreachableCode
        return None

    @classmethod
    def _resolve_context(
        cls, context: InputContractContext | InputEIP712Context, out: OutputAdder
    ) -> ResolvedContractContext | ResolvedEIP712Context | None:
        match context:
            case InputContractContext():
                return cls._resolve_context_contract(context, out)
            case InputEIP712Context():
                return cls._resolve_context_eip712(context, out)
            case _:
                assert_never(context)

    @classmethod
    def _resolve_metadata(cls, metadata: InputMetadata, out: OutputAdder) -> ResolvedMetadata | None:
        resolved_enums = {}
        if metadata.enums is not None:
            for enum_id, enum in metadata.enums.items():
                if (resolved_enum := cls._resolve_enum(enum, out)) is not None:
                    resolved_enums[enum_id] = resolved_enum

        return ResolvedMetadata(
            owner=metadata.owner,
            info=metadata.info,
            token=metadata.token,
            enums=resolved_enums,
        )

    @classmethod
    def _resolve_enum(cls, enum: HttpUrl | EnumDefinition, out: OutputAdder) -> dict[str, str] | None:
        match enum:
            case HttpUrl() as url:
                try:
                    return client.get(url=url, model=EnumDefinition)
                except Exception as e:
                    return out.error(
                        title="Failed to fetch enum definition from URL",
                        message=f'Failed to fetch enum definition from URL "{url}": {e}',
                    )
            case dict():
                return enum
            case _:
                assert_never(enum)

    @classmethod
    def _resolve_context_contract(
        cls, context: InputContractContext, out: OutputAdder
    ) -> ResolvedContractContext | None:
        if (contract := cls._resolve_contract(context.contract, out)) is None:
            return None

        return ResolvedContractContext.model_validate({"$id": context.id, "contract": contract})

    @classmethod
    def _resolve_contract(cls, contract: InputContract, out: OutputAdder) -> ResolvedContract | None:
        if (abi := cls._resolve_abis(contract.abi, out)) is None:
            return None
        if (deployments := cls._resolve_deployments(contract.deployments, out)) is None:
            return None

        if contract.factory is None:
            factory = None
        elif (factory := cls._resolve_factory(contract.factory, out)) is None:
            return None

        return ResolvedContract(
            abi=abi, deployments=deployments, addressMatcher=contract.addressMatcher, factory=factory
        )

    @classmethod
    def _resolve_deployments(
        cls, deployments: list[InputDeployment], out: OutputAdder
    ) -> list[ResolvedDeployment] | None:
        resolved_deployments = []
        for deployment in deployments:
            if (resolved_deployment := cls._resolve_deployment(deployment, out)) is not None:
                resolved_deployments.append(resolved_deployment)
        return resolved_deployments

    @classmethod
    def _resolve_deployment(cls, deployment: InputDeployment, out: OutputAdder) -> ResolvedDeployment | None:
        return ResolvedDeployment(chainId=deployment.chainId, address=Address(deployment.address))

    @classmethod
    def _resolve_factory(cls, factory: InputFactory, out: OutputAdder) -> ResolvedFactory | None:
        if (deployments := cls._resolve_deployments(factory.deployments, out)) is None:
            return None

        return ResolvedFactory(deployments=deployments, deployEvent=factory.deployEvent)

    @classmethod
    def _resolve_abis(cls, abis: list[ABI] | HttpUrl, out: OutputAdder) -> list[ABI] | None:
        match abis:
            case HttpUrl() as url:
                try:
                    return client.get(url=url, model=list[ABI])
                except Exception as e:
                    return out.error(
                        title="Failed to fetch ABI from URL",
                        message=f'Failed to fetch ABI from URL "{url}": {e}',
                    )
            case list():
                return abis
            case _:
                assert_never(abis)

    @classmethod
    def _resolve_context_eip712(cls, context: InputEIP712Context, out: OutputAdder) -> ResolvedEIP712Context | None:
        if (eip712 := cls._resolve_eip712(context.eip712, out)) is None:
            return None

        return ResolvedEIP712Context.model_validate({"$id": context.id, "eip712": eip712})

    @classmethod
    def _resolve_eip712(cls, eip712: InputEIP712, out: OutputAdder) -> ResolvedEIP712 | None:
        if eip712.domain is None:
            domain = None
        elif (domain := cls._resolve_domain(eip712.domain, out)) is None:
            return None

        if (schemas := cls._resolve_schemas(eip712.schemas, out)) is None:
            return None
        if (deployments := cls._resolve_deployments(eip712.deployments, out)) is None:
            return None

        return ResolvedEIP712(
            domain=domain,
            schemas=schemas,
            domainSeparator=eip712.domainSeparator,
            deployments=deployments,
        )

    @classmethod
    def _resolve_domain(cls, domain: InputDomain, out: OutputAdder) -> ResolvedDomain | None:
        return ResolvedDomain(
            name=domain.name,
            version=domain.version,
            chainId=domain.chainId,
            verifyingContract=None if domain.verifyingContract is None else Address(domain.verifyingContract),
        )

    @classmethod
    def _resolve_schemas(cls, schemas: list[EIP712Schema | HttpUrl], out: OutputAdder) -> list[EIP712Schema] | None:
        resolved_schemas = []
        for schema in schemas:
            if (resolved_schema := cls._resolve_schema(schema, out)) is not None:
                resolved_schemas.append(resolved_schema)
        return resolved_schemas

    @classmethod
    def _resolve_schema(cls, schema: EIP712Schema | HttpUrl, out: OutputAdder) -> EIP712Schema | None:
        match schema:
            case HttpUrl() as url:
                try:
                    return client.get(url=url, model=EIP712Schema)
                except Exception as e:
                    return out.error(
                        title="Failed to fetch EIP-712 schema from URL",
                        message=f'Failed to fetch EIP-712 schema from URL "{url}": {e}',
                    )
            case EIP712Schema():
                return schema
            case _:
                assert_never(schema)

    @classmethod
    def _resolve_display(
        cls,
        display: InputDisplay,
        context: ResolvedContractContext | ResolvedEIP712Context,
        enums: dict[Id, EnumDefinition] | None,
        constants: ConstantProvider,
        out: OutputAdder,
    ) -> ResolvedDisplay | None:
        definitions = display.definitions or {}
        enums = enums or {}
        formats = {}
        for format_id, format in display.formats.items():
            if (resolved_format_id := cls._resolve_format_id(format_id, context, out)) is None:
                return None
            if (resolved_format := cls._resolve_format(format, definitions, enums, constants, out)) is None:
                return None
            if resolved_format_id in formats:
                return out.error(
                    title="Duplicate format",
                    message=f"Descriptor contains 2 formats sections for {resolved_format_id}",
                )
            formats[resolved_format_id] = resolved_format

        return ResolvedDisplay(formats=formats)

    @classmethod
    def _resolve_field_description(
        cls,
        prefix: DataPath,
        definition: InputFieldDescription,
        enums: dict[Id, EnumDefinition],
        constants: ConstantProvider,
        out: OutputAdder,
    ) -> ResolvedFieldDescription | None:
        match definition.format:
            case None | FieldFormat.RAW | FieldFormat.AMOUNT | FieldFormat.TOKEN_AMOUNT | FieldFormat.DURATION:
                pass
            case (
                FieldFormat.ADDRESS_NAME
                | FieldFormat.CALL_DATA
                | FieldFormat.NFT_NAME
                | FieldFormat.DATE
                | FieldFormat.UNIT
                | FieldFormat.ENUM
            ):
                if definition.params is None:
                    return out.error(
                        title="Missing parameters",
                        message=f"""Field format "{definition.format.value}" requires parameters to be defined, """
                        f"""they are missing for field "{definition.path}".""",
                    )
            case _:
                assert_never(definition.format)

        params = resolve_field_parameters(prefix, definition.params, enums, constants, out)

        if (value := resolve_field_value(prefix, definition, definition.format, constants, out)) is None:
            return None

        return ResolvedFieldDescription.model_validate(
            {
                "$id": definition.id,
                "value": value,
                "label": constants.resolve(definition.label, out),
                "format": FieldFormat(definition.format) if definition.format is not None else None,
                "params": params,
            }
        )

    @classmethod
    def _resolve_format_id(
        cls,
        format_id: str,
        context: ResolvedContractContext | ResolvedEIP712Context,
        out: OutputAdder,
    ) -> EIP712Type | Selector | None:
        match context:
            case ResolvedContractContext():
                if format_id.startswith("0x"):
                    return Selector(format_id)

                if (reduced_signature := reduce_signature(format_id)) is not None:
                    return Selector(signature_to_selector(reduced_signature))

                return out.error(
                    title="Invalid selector",
                    message=f""""{format_id}" is not a valid function signature or selector.""",
                )
            case ResolvedEIP712Context():
                return format_id
            case _:
                assert_never(context)

    @classmethod
    def _resolve_format(
        cls,
        format: InputFormat,
        definitions: dict[Id, InputFieldDefinition],
        enums: dict[Id, EnumDefinition],
        constants: ConstantProvider,
        out: OutputAdder,
    ) -> ResolvedFormat | None:
        if (fields := cls._resolve_fields(ROOT_DATA_PATH, format.fields, definitions, enums, constants, out)) is None:
            return None

        return ResolvedFormat.model_validate(
            {
                "$id": format.id,
                "intent": format.intent,
                "fields": fields,
                "required": format.required,
                "excluded": format.excluded,
                "screens": format.screens,
            }
        )

    @classmethod
    def _resolve_fields(
        cls,
        prefix: DataPath,
        fields: list[InputField],
        definitions: dict[Id, InputFieldDefinition],
        enums: dict[Id, EnumDefinition],
        constants: ConstantProvider,
        out: OutputAdder,
    ) -> list[ResolvedField] | None:
        resolved_fields = []
        for input_format in fields:
            if (resolved_field := cls._resolve_field(prefix, input_format, definitions, enums, constants, out)) is None:
                return None
            resolved_fields.extend(resolved_field)
        return resolved_fields

    @classmethod
    def _resolve_field(
        cls,
        prefix: DataPath,
        field: InputField,
        definitions: dict[Id, InputFieldDefinition],
        enums: dict[Id, EnumDefinition],
        constants: ConstantProvider,
        out: OutputAdder,
    ) -> list[ResolvedField] | None:
        resolved_fields: list[ResolvedField] = []
        match field:
            case InputReference():
                if (resolved_field := resolve_reference(prefix, field, definitions, enums, constants, out)) is None:
                    return None
                resolved_fields.append(resolved_field)
            case InputFieldDescription():
                if (resolved_field := cls._resolve_field_description(prefix, field, enums, constants, out)) is None:
                    return None
                resolved_fields.append(resolved_field)
            case InputNestedFields():
                if (
                    resolved_nested_fields := cls._resolve_nested_fields(
                        prefix, field, definitions, enums, constants, out
                    )
                ) is None:
                    return None
                resolved_fields.extend(resolved_nested_fields)
            case _:
                assert_never(field)
        return resolved_fields

    @classmethod
    def _resolve_nested_fields(
        cls,
        prefix: DataPath,
        fields: InputNestedFields,
        definitions: dict[Id, InputFieldDefinition],
        enums: dict[Id, EnumDefinition],
        constants: ConstantProvider,
        out: OutputAdder,
    ) -> list[ResolvedNestedFields | ResolvedFieldDescription] | None:
        if fields.path is None:
            return out.error(
                title="Unsupported nested fields value",
                message="Nested fields are only supported with data paths and not constant values.",
            )

        path: DataPath
        match constants.resolve_path(fields.path, out):
            case None:
                return None
            case DataPath() as data_path:
                path = data_path_concat(prefix, data_path)
            case ContainerPath() as container_path:
                return out.error(
                    title="Invalid path type",
                    message=f"Container path {container_path} cannot be used with nested fields.",
                )
            case _:
                assert_never(fields.path)

        if (
            resolved_fields := cls._resolve_fields(
                prefix=path, fields=fields.fields, definitions=definitions, enums=enums, constants=constants, out=out
            )
        ) is None:
            return None

        match path.elements[-1]:
            case Field() | ArrayElement():
                return resolved_fields
            case ArraySlice():
                return out.error(
                    title="Invalid nested fields",
                    message="Using nested fields on an array slice is not allowed.",
                )
            case Array():
                return [ResolvedNestedFields(value=ResolvedValuePath(path=path), fields=resolved_fields)]
            case _:
                assert_never(path.elements[-1])



import pytest
from typing_extensions import TypeVar

_T = TypeVar("_T")


def single_or_skip(value: _T | dict[str, _T] | None) -> _T:
    assert value is not None
    if isinstance(value, dict):
        pytest.skip("Multiple descriptors tests not supported")
    return value


def single_or_first(value: _T | dict[str, _T] | None) -> _T:
    assert value is not None
    if isinstance(value, dict):
        return next(iter(value.values()))
    return value



import json
import os
from collections.abc import Iterator
from json import JSONEncoder
from pathlib import Path
from typing import Any, override


def read_jsons_with_includes(paths: list[Path]) -> Any:
    """
    Read a list JSON files, recursively inlining any included files.

    Keys from the calling file override those of the included file (the included file defines defaults).
    Keys from the later files in the list override those of the first files.

    Note:
      - circular includes are not detected and will result in a stack overflow.
      - "includes" key can only be used at root level of an object.
    """
    result: dict[str, Any] = {}
    for path in paths:
        # noinspection PyTypeChecker
        result = _merge_dicts(result, read_json_with_includes(path))
    return result


def read_json_with_includes(path: Path) -> Any:
    """
    Read a JSON file, recursively inlining any included files.

    Keys from the calling file override those of the included file (the included file defines defaults).

    If include is a list, files are included in other they are defined, with later files overriding previous files.

    Note:
      - circular includes are not detected and will result in a stack overflow.
      - "includes" key can only be used at root level of an object.
    """
    result: dict[str, Any] = dict_from_json_file(path)
    if isinstance(result, dict) and (includes := result.pop("includes", None)) is not None:
        if isinstance(includes, list):
            parent = read_jsons_with_includes(paths=[path.parent / p for p in includes])
        else:
            # noinspection PyTypeChecker
            parent = read_json_with_includes(path.parent / includes)
        result = _merge_dicts(parent, result)
    return result


def _merge_dicts(d1: dict[str, Any], d2: dict[str, Any]) -> dict[str, Any]:
    """
    Merge d1 and d2, with priority to d2.

    Recursively called when dicts are encountered.

    This function assumes that if the same field is in both dicts, then the types must be the same.
    """
    merged = {}
    for key, val1 in d1.items():
        if (val2 := d2.get(key)) is not None:
            if isinstance(val1, dict):
                merged[key] = _merge_dicts(d1=val1, d2=val2)
            else:
                merged[key] = val2
        else:
            merged[key] = val1
    return {**d2, **merged}


def dict_from_json_str(value: str) -> dict[str, Any]:
    """Deserialize a dict from a JSON string."""
    return json.loads(value)


def dict_from_json_file(path: Path) -> dict[str, Any]:
    """Deserialize a dict from a JSON file."""
    with open(path, "rb") as f:
        return json.load(f)


def dict_to_json_str(values: dict[str, Any]) -> str:
    """Serialize a dict into a JSON string."""
    return json.dumps(values, indent=2, cls=CompactJSONEncoder)


def dict_to_json_file(path: Path, values: dict[str, Any]) -> None:
    """Serialize a dict into a JSON file, creating parent directories as needed."""
    os.makedirs(path.parent, exist_ok=True)
    with open(path, "w") as f:
        f.write(dict_to_json_str(values))
        f.write("\n")


class CompactJSONEncoder(JSONEncoder):
    """A JSON Encoder that puts small containers on single lines."""

    CONTAINER_TYPES = (list, tuple, dict)
    """Container datatypes include primitives or other containers."""

    MAX_WIDTH = 120
    """Maximum width of a container that might be put on a single line."""

    MAX_ITEMS = 10
    """Maximum number of items in container that might be put on single line."""

    PRIMITIVES_ONLY = False
    """Only put containers containing primitives only on a single line."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if kwargs.get("indent") is None:
            kwargs["indent"] = 2
        super().__init__(*args, **kwargs)
        self.indentation_level = 0

    @override
    def encode(self, o: Any) -> str:
        if isinstance(o, list | tuple):
            return self._encode_list(o)
        if isinstance(o, dict):
            return self._encode_object(o)
        if isinstance(o, float):  # Use scientific notation for floats
            return format(o, "g")
        return json.dumps(
            obj=o,
            skipkeys=self.skipkeys,
            ensure_ascii=self.ensure_ascii,
            check_circular=self.check_circular,
            allow_nan=self.allow_nan,
            sort_keys=self.sort_keys,
            indent=self.indent,
            separators=(self.item_separator, self.key_separator),
            default=self.default if hasattr(self, "default") else None,
        )

    def _encode_list(self, o: list[Any] | tuple[Any]) -> str:
        if self._put_on_single_line(o):
            return "[" + ", ".join(self.encode(el) for el in o) + "]"
        self.indentation_level += 1
        output = [self.indent_str + self.encode(el) for el in o]
        self.indentation_level -= 1
        return "[\n" + ",\n".join(output) + "\n" + self.indent_str + "]"

    def _encode_object(self, o: Any) -> str:
        if not o:
            return "{}"

        o = {str(k) if k is not None else "null": v for k, v in o.items()}

        if self.sort_keys:
            o = dict(sorted(o.items(), key=lambda x: x[0]))

        if self._put_on_single_line(o):
            return "{ " + ", ".join(f"{json.dumps(k)}: {self.encode(el)}" for k, el in o.items()) + " }"

        self.indentation_level += 1
        output = [f"{self.indent_str}{json.dumps(k)}: {self.encode(v)}" for k, v in o.items()]
        self.indentation_level -= 1

        return "{\n" + ",\n".join(output) + "\n" + self.indent_str + "}"

    @override
    def iterencode(self, o: Any, _one_shot: bool = False) -> Iterator[str]:
        return self.encode(o)  # type: ignore

    def _put_on_single_line(self, o: Any) -> bool:
        return self._primitives_only(o) and len(o) <= self.MAX_ITEMS and len(str(o)) - 2 <= self.MAX_WIDTH

    def _primitives_only(self, o: list[Any] | tuple[Any] | dict[Any, Any]) -> bool:
        if not self.PRIMITIVES_ONLY:
            return True
        if isinstance(o, list | tuple):
            return not any(isinstance(el, self.CONTAINER_TYPES) for el in o)
        elif isinstance(o, dict):
            return not any(isinstance(el, self.CONTAINER_TYPES) for el in o.values())

    @property
    def indent_str(self) -> str:
        if isinstance(self.indent, int):
            return " " * (self.indentation_level * self.indent)
        elif isinstance(self.indent, str):
            return self.indentation_level * self.indent
        else:
            raise ValueError(f"indent must either be of type int or str (is: {type(self.indent)})")



from dataclasses import dataclass
from pathlib import Path

from tests.files import ERC7730_REGISTRY, LEGACY_REGISTRY, PROJECT_ROOT


@dataclass(frozen=True, kw_only=True)
class TestCase:
    """Define a test case."""

    id: str
    label: str
    description: str
    error: str | None = None


def case_id(case: TestCase) -> str:
    """Generate test case identifier for a TestCase."""
    return case.id


def path_id(path: Path) -> str:
    """Generate test case identifier for a path."""
    for base in (ERC7730_REGISTRY, LEGACY_REGISTRY):
        if base in path.parents:
            return str(path.relative_to(base))
    return str(path.relative_to(PROJECT_ROOT))



import pytest

from erc7730.common.abi import ABIDataType
from erc7730.common.output import RaisingOutputAdder
from erc7730.convert.resolved.values import encode_value
from erc7730.model.types import HexStr, ScalarType


@pytest.mark.parametrize(
    "abi_type,value,expected",
    [
        (ABIDataType.UINT, "0xaaa", "0xaaa"),
        (ABIDataType.INT, "0xaaa", "0xaaa"),
        (ABIDataType.UFIXED, "0xaaa", "0xaaa"),
        (ABIDataType.FIXED, "0xaaa", "0xaaa"),
        (ABIDataType.ADDRESS, "0xaaa", "0xaaa"),
        (ABIDataType.BOOL, "0xaaa", "0xaaa"),
        (ABIDataType.BYTES, "0xaaa", "0xaaa"),
        (ABIDataType.STRING, "0xaaa", "0xaaa"),
        (ABIDataType.UINT, 42, "0x2a"),
        (ABIDataType.INT, 42, "0x2a"),
        (ABIDataType.INT, -42, "0xd6"),
        (ABIDataType.BOOL, True, "0x01"),
        (ABIDataType.BOOL, False, "0x00"),
        (ABIDataType.BYTES, "0xcafebabe", "0xcafebabe"),
        (
            ABIDataType.BYTES,
            "0xcafebabecafebabecafebabecafebabecafebabecafebabe",
            "0xcafebabecafebabecafebabecafebabecafebabecafebabe",
        ),
        (
            ABIDataType.ADDRESS,
            "0x11111112542D85B3EF69AE05771c2dCCff4fAa26",
            "0x11111112542d85b3ef69ae05771c2dccff4faa26",
        ),
        (ABIDataType.STRING, "hi", "0x6869"),
        (
            ABIDataType.STRING,
            "hi this is a very long string, because we want to test what happens with several chunks",
            "0x6869207468697320697320612076657279206c6f6e6720737472696e672c20626563617573652077652077616e7420746f20746"
            "5737420776861742068617070656e732077697468207365766572616c206368756e6b73",
        ),
    ],
)
def test_encode_value(abi_type: ABIDataType, value: ScalarType, expected: HexStr) -> None:
    assert encode_value(value, abi_type, RaisingOutputAdder()) == expected


@pytest.mark.parametrize(
    "abi_type,value,expected",
    [
        (ABIDataType.UINT, -42, """Value "-42" is not an unsigned int"""),
        (ABIDataType.UINT, "42", """Value "42" is not an unsigned int"""),
        (ABIDataType.ADDRESS, "42", """Value "42" is not a valid address hexadecimal string"""),
        (ABIDataType.BOOL, "42", """Value "42" is not a bool"""),
        (ABIDataType.BYTES, "42", """Value "42" is not a valid hexadecimal string"""),
        (ABIDataType.STRING, 42, """Value "42" is not a string"""),
        (ABIDataType.FIXED, 42, """Fixed precision numbers are not supported"""),
        (ABIDataType.UFIXED, 42, """Fixed precision numbers are not supported"""),
    ],
)
def test_encode_value_error(abi_type: ABIDataType, value: ScalarType, expected: str) -> None:
    with pytest.raises(Exception) as exc_info:
        encode_value(value, abi_type, RaisingOutputAdder())
    assert expected in str(exc_info.value)



"""
Package implementing an object model for ERC-7730 input descriptors.

This model represents descriptors before resolution phase:
    - URLs have been not been fetched yet
    - Contract addresses have not been normalized to lowercase
    - References have not been inlined
    - Constants have not been inlined
    - Field definitions have not been inlined
    - Nested fields have been flattened where possible
    - Selectors have been converted to 4 bytes form
"""

from pydantic import Field

from erc7730.model.base import Model
from erc7730.model.input.context import InputContractContext, InputEIP712Context
from erc7730.model.input.display import InputDisplay
from erc7730.model.input.metadata import InputMetadata

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class InputERC7730Descriptor(Model):
    """
    An ERC7730 Clear Signing descriptor.

    This model is directly serializable back the original JSON document.

    Specification: https://github.com/LedgerHQ/clear-signing-erc7730-registry/tree/master/specs

    JSON schema: https://github.com/LedgerHQ/clear-signing-erc7730-registry/blob/master/specs/erc7730-v1.schema.json
    """

    schema_: str | None = Field(
        None,
        alias="$schema",
        description="The schema that the document should conform to. This should be the URL of a version of the clear "
        "signing JSON schemas available under "
        "https://github.com/LedgerHQ/clear-signing-erc7730-registry/tree/master/specs",
    )

    context: InputContractContext | InputEIP712Context = Field(
        title="Binding Context Section",
        description="The binding context is a set of constraints that are used to bind the ERC7730 file to a specific"
        "structured data being displayed. Currently, supported contexts include contract-specific"
        "constraints or EIP712 message specific constraints.",
    )

    metadata: InputMetadata = Field(
        title="Metadata Section",
        description="The metadata section contains information about constant values relevant in the scope of the"
        "current contract / message (as matched by the `context` section)",
    )

    display: InputDisplay = Field(
        title="Display Formatting Info Section",
        description="The display section contains all the information needed to format the data in a human readable"
        "way. It contains the constants and formatters used to display the data contained in the bound structure.",
    )



from pathlib import Path

import pytest

from erc7730.lint.lint import lint_all_and_print_errors
from tests.cases import path_id
from tests.files import ERC7730_DESCRIPTORS


@pytest.mark.parametrize("input_file", ERC7730_DESCRIPTORS, ids=path_id)
def test_registry_files(input_file: Path) -> None:
    """
    Test linting ERC-7730 registry files, which should all be valid at all times.
    """
    assert lint_all_and_print_errors([input_file])



"""
Base types for ERC-7730 descriptors.

Specification: https://github.com/LedgerHQ/clear-signing-erc7730-registry/tree/master/specs
JSON schema: https://github.com/LedgerHQ/clear-signing-erc7730-registry/blob/master/specs/erc7730-v1.schema.json
"""

from typing import Annotated

from pydantic import BeforeValidator, Field

from erc7730.common.pydantic import ErrorTypeLabel

Id = Annotated[
    str,
    Field(
        title="Id",
        description="An internal identifier that can be used either for clarity specifying what the element is or as a "
        "reference in device specific sections.",
        min_length=1,
        examples=["some_identifier"],
    ),
    ErrorTypeLabel('identifier, such as "some_identifier".'),
]

MixedCaseAddress = Annotated[
    str,
    Field(
        title="Contract Address",
        description="An Ethereum contract address, can be lowercase or EIP-55.",
        min_length=42,
        max_length=42,
        pattern=r"^0x[a-fA-F0-9]+$",
    ),
    ErrorTypeLabel(
        '20 bytes, hexadecimal Ethereum address prefixed with "0x" (EIP-55 or lowercase), such as '
        + '"0xdac17f958d2ee523a2206206994597c13d831ec7".'
    ),
]

Address = Annotated[
    str,
    Field(
        title="Contract Address",
        description="An Ethereum contract address (normalized to lowercase).",
        min_length=42,
        max_length=42,
        pattern=r"^0x[a-f0-9]+$",
    ),
    BeforeValidator(lambda v: v.lower()),
    ErrorTypeLabel(
        '20 bytes, lowercase hexadecimal Ethereum address prefixed with "0x", such as '
        + '"0xdac17f958d2ee523a2206206994597c13d831ec7".'
    ),
]

Selector = Annotated[
    str,
    Field(
        title="Selector",
        description="An Ethereum contract function identifier, in 4 bytes, hex encoded form.",
        min_length=10,
        max_length=10,
        pattern=r"^0x[a-z0-9]+$",
    ),
    ErrorTypeLabel(
        '4 bytes, lowercase hexadecimal Ethereum function selector prefixed with "0x", such as ' + '"0xdac17f95".'
    ),
]

HexStr = Annotated[
    str,
    Field(
        title="Hexadecimal string",
        description="A byte array encoded as an hexadecimal string.",
        pattern=r"^0x[a-f0-9]+$",
    ),
    BeforeValidator(lambda v: v.lower()),
    ErrorTypeLabel('lowercase hexadecimal string prefixed with "0x", such as "0xdac17f95".'),
]

ScalarType = str | int | bool | float
